import os
import sys
import copy
import shutil
import pickle
import random
import subprocess
import numpy as np
import multiprocessing as mp
from tqdm import tqdm

from datetime import datetime
from itertools import combinations, chain
from scipy.sparse import lil_matrix, csr_matrix, save_npz, load_npz, triu
from collections import defaultdict, Counter
from mycotools.lib.biotools import gff2list, dict2fa
from mycotools.lib.kontools import write_json, read_json, collect_files, \
                                   checkdir, eprint
from cloci.lib import input_parsing, treecalcs, evo_conco, output_data

try:
    from graph_tool.all import *
except ImportError:
    eprint('\nERROR: graph-tool is not installed. Please install via conda', flush = True)
    sys.exit(147)

def hash_hgx(gff_path, ome, hgx_genes, gene2hg, clusplusminus):
    """sliding window of size clusplusminus, identify sequences that may have
    hgs that may be significant, grab their window, and check for a significant
    higher order og combo in the locus"""
        
    gff_list, gene_dict = gff2list(gff_path), defaultdict(list)
    cds_dict = input_parsing.compileCDS(gff_list, os.path.basename(gff_path).replace('.gff3',''))
                
    hgx_dict = {}   
    for scaf in cds_dict: 
        for i0, seq0 in enumerate(cds_dict[scaf]):
            if seq0 in hgx_genes: # if the og is part of a significant seed
            # locus 
                if i0 < clusplusminus: # is i0 - clusplusminus < 0 ?
                    locus = cds_dict[scaf][:i0+clusplusminus+1] # then gather
                    # all the beginning
                else:
                    locus = cds_dict[scaf][i0-clusplusminus:i0+clusplusminus+1]
                    # instead get the +/- and adjust for python
                og0 = gene2hg[seq0]
                for hgx in hgx_genes[seq0]:
                    if hgx not in hgx_dict:
                        hgx_dict[hgx] = {og: [] for og in hgx}
                    try:
                        hgx_dict[hgx][og0].append(seq0) # add the sequence to the hgx_dict
                    except KeyError:
                        pass
                    start, end = None, None
                    for i1, seq1 in enumerate(locus): # for each index and sequence
                    # in the locus
                        try:
                            og1 = gene2hg[seq1]
                        except KeyError: # missing entry
                            continue
                        if og1 in set(hgx) and start is None: # if the og is
                        # in the sig clus and we haven't started
                            start = i1 # this is the start index
                            hgx_dict[hgx][og1].append(seq1)

                        elif og1 in set(hgx): # otherwise if it is in the
                        # sig clus label this the other border unless another is 
                        # found
                            end = i1 + 1
                            hgx_dict[hgx][og1].append(seq1)
                    for gene in locus[start:end]:
                        gene_dict[gene].append(hgx)

    gene_tup = tuple([
        tuple([gene, tuple(sorted(vals))]) \
        for gene, vals in gene_dict.items()
        ])
    # ((gene, {gene: [HGx...]}))

    hgx_tup = tuple([
        tuple([
            hgx,
            tuple([tuple([og, tuple(sorted(set(seqs)))]) \
                for og, seqs in hgs.items()]) \
            ])
            for hgx, hgs in hgx_dict.items()
        ])

    return ome, gene_tup, hgx_tup


def find_hgx_pairs(gene2hgx, ome, hgx2i, pairsDict, min_hgx_overlap = 2):
    # is this going to merge hgxs that overlap by one gene in min_hgx_overlap loci?
    hgx_pairs_raw = []
    # check for gene overlap in hgxs
    for hgxs in list(gene2hgx.values()):
        # add each possible hgxpair
        hgx_pairs_raw.extend(list(combinations(sorted(hgxs), 2)))

    hgx_pairs_count = [tuple((hgx2i[y] for y in x)) \
                       for x, g in Counter(hgx_pairs_raw).items() \
                       if g >= min_hgx_overlap]
    for hgxpair in hgx_pairs_count:
        pairsDict[hgxpair].append(ome)
    return pairsDict


def read_mci_rows(rows_file):
    mci2pre = {}
    with open(rows_file, 'r') as raw:
        for line in raw:
            mci, pre = [int(x) for x in line.rstrip().split()]
            mci2pre[mci] = pre
    return mci2pre


def MCL(adj_path, clus_file, inflation = 1.5, threads = 1, rows_file = None):

    mci_file = adj_path[:-4] + '.mci'
    if not rows_file:
        rows_file = os.path.dirname(adj_path) + '/mcl_rows.tsv'
    if not os.path.isfile(mci_file):
        subprocess.call(['mcxload', '-abc', adj_path, '-o',
                         mci_file + '.tmp', '--write-binary',
                         '-write-tab', rows_file, '--stream-mirror'],
                         stdout = subprocess.DEVNULL, 
                         stderr = subprocess.DEVNULL)
        shutil.move(mci_file + '.tmp', mci_file)
    subprocess.call([
        'mcl', mci_file, '-I', str(inflation),
        '-o', clus_file + '.tmp', '-te', str(threads)
        ], stdout = subprocess.DEVNULL, stderr = subprocess.DEVNULL
        )
    subprocess.call(['mcxdump', '-icl', clus_file + '.tmp',
                     '-o', clus_file, '--dump-pairs'],
                    stdout = subprocess.DEVNULL, stderr = subprocess.DEVNULL)
    os.remove(clus_file + '.tmp')


def sorensen(interlen, len0, len1):
    return 2*interlen/(len0 + len1)
def jaccard(interlen, len0, len1):
    return interlen/(len0 + len1 - interlen)
def overlap(interlen, len0, len1):
    return interlen/min([len0, len1])

def calc_loc2loc_sim_mat(
    i0, i1, index, loc0, loc1, hgL0, hgL1, coef, blast_ids,
    min_hlg_id = 0.15
    ):

    # Identify the overlapping HGs
    hg_dict = defaultdict(lambda: [[], []])
    for i, hg in enumerate(hgL0):
        hg_dict[hg][0].append(loc0[i])
    for i, hg in enumerate(hgL1):
        hg_dict[hg][1].append(loc1[i])
    hg_dict = {k: v for k, v in hg_dict.items() \
               if v[0] and v[1]}

    scores = []
    for hg, gene_lists in hg_dict.items():
        scores.append([])
        # for each pairwise gene comparison
        for gene0 in gene_lists[0]:
            for gene1 in gene_lists[1]:
                try:
                    scores[-1].append(blast_ids[hg][gene0][gene1])
                # if the genes are missing its a 0
                # the alignment algorithm MUST output enough alignments
                # for this is to reliably work on widely dispersed clusters
                except KeyError: 
                    pass
    scores = [x for x in scores if x]
    if scores:
        maxScores = [max(i) for i in scores]
        hlg_id = (sum(maxScores)/len(maxScores)) * coef
        if hlg_id > min_hlg_id:
            return f'{i0 + index}\t{i1 + index}\t{hlg_id}\n'
        else:
            return ''



def acquire_clus_hlg_sim_noq(
    i0, i1, index, loc0, loc1, hgL0, hgL1, set0, set1, blast_ids,
    simfun = overlap, min_hlg_id = 0.15
    ):

    # Identify the overlapping HGs
    intersection = set0.intersection(set1)
    inter_list = list(intersection)
    hg_dict = {k: [[], []] for k in inter_list}

    # populate an hg2gene dict for ome0 and ome1
    for i, hg in enumerate(hgL0):
        if hg in intersection:
            hg_dict[hg][0].append(loc0[i])
    for i, hg in enumerate(hgL1):
        if hg in intersection:
            hg_dict[hg][1].append(loc1[i])

    scores = []
    for hg in inter_list:
        # check to verify the genes are recovered
        scores.append([])
        # for each pairwise gene comparison
        for gene0 in hg_dict[hg][0]:
            for gene1 in hg_dict[hg][1]:
                try:
                    scores[-1].append(blast_ids[hg][gene0][gene1])
                # if the genes are missing its a 0
                # the alignment algorithm MUST output enough alignments
                # for this is to reliably work on widely dispersed clusters
                except KeyError: 
                    scores[-1].append(0)
    scores = [x for x in scores if x]
    if scores:
        maxScores = [max(i) for i in scores]
        coef = simfun(len(scores), len(set0), len(set1))
        hlg_id = (sum(maxScores)/len(maxScores)) * coef
        if hlg_id > min_hlg_id:
            return f'{i0 + index}\t{i1 + index}\t{hlg_id}\n'
        else:
            return ''


def rnd2_gen_blastids_mp(hg, group_dict, hgx_dir, grp_dir, minid):#, blast_ids_mngr):
    gene_set = set(chain(*list(group_dict.values())))
    gene_len = len(gene_set)
    algn_base = f'{hgx_dir}{hg}.out'
    json_out = f'{grp_dir}/algn/{hg}.json'
    if not os.path.isfile(json_out):
        try:
            hg_dict = parse_blast(algn_base, gene_len, gene_set, minid)
            new_hg_dict = defaultdict(dict)
            for g, genes in group_dict.items():
                for g0, g1 in combinations(sorted(genes), 2):
                    try:
                        new_hg_dict[g0][g1] = hg_dict[g0][g1]
                    except KeyError:
                        pass
                    try:
                        new_hg_dict[g1][g0] = hg_dict[g1][g0] 
                    except KeyError:
                        pass
            write_json(new_hg_dict, json_out + '.tmp')
            os.rename(json_out + '.tmp', json_out)
        except FileNotFoundError:
            pass



def gen_blastids_mp(hg, genes, hgx_dir, grp_dir, minid):#, blast_ids_mngr):
    # CURRENTLY NOT SETUP FOR ANY OTHER CLANS THAN 0
    # would need to incorporate clanI in the output json
    # or else it runs the risk of using a failed runs' old jsons
    # cannot multiprocess the runs either
    gene_set = set(genes)
    gene_len = len(gene_set)
    algn_base = f'{hgx_dir}{hg}.out'
    json_out = f'{grp_dir}/algn/{hg}.json'
    if not os.path.isfile(json_out):
        try:
            hg_dict = parse_blast(algn_base, gene_len, gene_set, minid)
            if hg_dict is not None:
                write_json(hg_dict, json_out + '.tmp')
                os.rename(json_out + '.tmp', json_out)
    #        blast_ids_mngr[hg].update(hg_dict)
        except FileNotFoundError:
            pass
 #       print(f'\t\t\t\t\tAttempting to BLAST {hg}', flush = True)
#        blast = run_hgx_blast(hg_dir, hg, genes, hgx_dir, 
  #                    blastp = 'blastp', cpus = 1)
   #     if blast:
    #        blast_ids_mngr[hg].update(parse_blast(algn_base, 
     #                                             gene_len, gene_set))


def read_hlg_sim_jsons(hlg_dir):
    jsons = collect_files(hlg_dir, 'json')
    blast_ids = {}
    for j_f in jsons:
        hg = os.path.basename(j_f).replace('.json','')
        blast_ids[int(hg)] = read_json(j_f)
    return blast_ids


def parse_blast(algn_base, gene_len, gene_set, minid = 30):
    hg_dict = defaultdict(dict)
    with open(algn_base, 'r') as raw:
        for line in raw:
            d = line.rstrip().split()
            # adjust to just reference d when safe
            try:
                q, s, pident = d[0], d[1], d[-1]
            except IndexError:
                eprint(f'ERROR: malformatted alignment: {os.path.basename(algn_base)}', 
                       flush = True) 
                return False
            if {q, s}.issubset(gene_set):
                # NEED a no self hits option in mmseqs or blast to avoid
                if q != s:
                    pid = float(pident)
                    if pid >= minid:
                        hg_dict[q][s] = pid/100 # adjust diamond to decimal
            if len(hg_dict) == gene_len:
                break
    return hg_dict


def hlg_sim_write_mngr(Q, sim_file):
    x = Q.get()
    with open(sim_file, 'a') as out:
        while x:
            out.write(x)
            x = Q.get()
            out.flush()


def calc_sim_by_row(adj_mat, hgLoci, loci, blast_ids, 
                    index, Q, max_complete, min_hlg_id, simfun):
 #   adj_mat = np.load(f'{mat_dir}{rowI}.npz')
    for ti0, overlap_loci in enumerate(adj_mat):
        i0 = ti0 + max_complete
        loc0, hgL0 = loci[i0], hgLoci[i0]
        sHGl0 = set([x for x in hgL0 if x is not None])
        for ti1 in overlap_loci:
            if ti1 > ti0:
                i1 = ti1 + max_complete
                loc1, hgL1 = loci[i1], hgLoci[i1]
                sHGl1 = set([x for x in hgL1 if x is not None])
                data = acquire_clus_hlg_sim_noq(i0, i1, index, loc0,
                                                loc1, hgL0, hgL1,
                                                sHGl0, sHGl1, blast_ids,
                                                min_hlg_id = min_hlg_id,
                                                simfun = simfun)
                if data:
                    Q.put(data)



def hlg_sim_mp_mngr(adj_mat, sim_file, max_complete, loci, hgLoci, index, 
                    blast_ids, hlg_dir, mat_dir, min_hlg_id, simfun, 
                    chunkscale = 100, cpus = 1):
    if not os.path.isdir(mat_dir):
        os.mkdir(mat_dir)
    
    # break up the adj_mat by chunksize and offload to file while preparing cmds
    row_len = len(adj_mat)
    chunksize = round(row_len/chunkscale)
    cmds = []
    Q = mp.Manager().Queue()
    for i, row in enumerate(range(0, row_len, chunksize)):
        if len(adj_mat) >= row + chunksize:
            sub_mat = [list(adj_mat[x]) for x in range(row, row + chunksize)]
        else:
            sub_mat = [list(adj_mat[x]) for x in range(row, len(adj_mat))]
#        np.save(f'{mat_dir}{i}.npz', sub_mat)
        locIs = [x for x in list(set(chain(*sub_mat))) if x >= row]
        t_hg_loc = {locI: hgLoci[locI + max_complete] for locI in locIs}
        t_loc = {locI: loci[locI + max_complete] for locI in locIs}
        t_hgs = set(chain(*list(t_hg_loc.values())))
        if None in t_hgs:
            t_hgs.remove(None)
        t_ids = {hg: blast_ids[hg] for hg in list(t_hgs) if hg in blast_ids}
        cmds.append((sub_mat, t_hg_loc, t_loc, t_ids, 
                     index, Q, max_complete, min_hlg_id, simfun))

    write_proc = mp.Process(target=hlg_sim_write_mngr, args=(Q, sim_file))
    write_proc.start()

    with mp.get_context('forkserver').Pool(processes = cpus - 1) as pool:
        pool.starmap(calc_sim_by_row, tqdm(cmds, total = len(cmds)))
        pool.close()
        pool.join()

    Q.put(None)
    write_proc.close()
    write_proc.join()

def calc_overlap_matrix(adj_arr, clan_arr):
    #INCORRECT! DOESN'T CURRENTLY USE MINIMUM!!!!!!
    return adj_arr/clan_arr.sum(axis=1).T

def calc_sorensen_matrix_sparse(adj_arr, clan_arr):
    return 2 * adj_arr/(clan_arr.sum(axis=1) + clan_arr.sum(axis=1).T)

def calc_sorensen_array(adj_arr, clan_arr):
    return 2 * adj_arr/(clan_arr.sum(axis=1) + np.array([clan_arr.sum(axis=1)]).T)

def matrix_sim_mngr(s_mat, out, loci, hgLoci, max_complete,
                    blast_ids, index, min_hlg_id):
    for ti0, row in enumerate(s_mat):
        i0 = ti0 + max_complete
        loc0, hgL0 = loci[i0], hgLoci[i0]
        for ti1, v in enumerate(row):
            i1 = ti1 + max_complete
            loc1, hgL1 = loci[i1], hgLoci[i1]
            data = calc_loc2loc_sim_mat(i0, i1, index, loc0,
                                        loc1, hgL0, hgL1,
                                        v, blast_ids,
                                        min_hlg_id = min_hlg_id)
            if data:
                out.write(data)
        out.flush()


def rnd2_loc2loc_sim_mngr(clan_arr, finished_file, max_complete, loci, hgLoci,
                     index, blast_ids, min_hlg_id, simfun):
    adj_arr = np.triu(clan_arr @ clan_arr.T)
    adj_arr[adj_arr < 2] = 0
    s_mat = calc_sorensen_array(adj_arr, clan_arr)
    np.fill_diagonal(s_mat, 0)
    with open(f'{finished_file}.w', 'a') as out:
        matrix_sim_mngr(s_mat, out, loci,
                        hgLoci, max_complete, blast_ids, index, min_hlg_id)



def loc2loc_sim_mngr(clan_arr, finished_file, max_complete, loci, hgLoci,
                     index, blast_ids, min_hlg_id, simfun):
    clan_arr = clan_arr.tocsr()
    # a matrix representing the cardinality of the intersection of hg_locs
    adj_arr = csr_matrix(triu(clan_arr @ clan_arr.transpose()))
    # less than two gene overlap is set to 0, will bias against small clusters
    # only have two genes; may lead to fragmented network topology if clusters
    # tend to be that small. 
    adj_arr.data[adj_arr.data < 2] = 0

    hits = np.split(adj_arr.indices, adj_arr.indptr)[1:-1]
    with open(f'{finished_file}.w', 'a') as out:
        for ti0, overlap_loci in enumerate(hits):
            i0 = ti0 + max_complete
            loc0, hgL0 = loci[i0], hgLoci[i0]
            sHGl0 = set([x for x in hgL0 if x is not None])
            for ti1 in overlap_loci[:]:
                i1 = ti1 + max_complete
                loc1, hgL1 = loci[i1], hgLoci[i1]
                sHGl1 = set([x for x in hgL1 if x is not None])
                data = acquire_clus_hlg_sim_noq(i0, i1, index, loc0,
                                                loc1, hgL0, hgL1,
                                                sHGl0, sHGl1, blast_ids,
                                                min_hlg_id = min_hlg_id,
                                                simfun = simfun)
                if data: 
                    out.write(data)
            out.flush()

def open_prev_res(finished_file, index):
    # identify the last definitively known clan
    with open(f'{finished_file}.w1', 'w') as out:
        with open(f'{finished_file}.w', 'r') as raw:
            data, max_complete = [], 0
            for line in raw:
                d = line.rstrip().split()
                try:
                    if d[0] != data[-1][0]:
                        out.write('\n'.join([' '.join(x) for x in data]) \
                                + '\n')
                        out.flush()
                        max_complete = int(data[0][0]) - index
                        data = []
                except IndexError: # first line
                    pass
                data.append(d)
    shutil.move(f'{finished_file}.w1', f'{finished_file}.w')
    return max_complete

def rnd2_loc2loc_mngr(hlg_dir, loci, hg_loci,
                     hgx_dir, groups, minid, min_hlg_id, simfun, cpus = 1):

    finished_file = f'{hlg_dir}loci.adj.tmp'
    # is this clan already completed?
    if os.path.isfile(finished_file) or os.path.isfile(finished_file + '.r'):
        return
    # was this clan partially written?
    elif os.path.isfile(f'{finished_file}.w'):
        os.remove(f'{finished_file}.w')

    # only examine loci that have not been completed
    ghg2i = defaultdict(dict)
    blast_hash = defaultdict(lambda: defaultdict(list))
    for g,v in enumerate(groups):
        i0, i1 = v
        for i2 in range(i0, i1):
            locus = loci[i2]
            hgs = hg_loci[i2]
            for i4, hg in enumerate(hgs):
                if hg is not None:
                    blast_hash[hg][g].append(locus[i4])
                    if hg not in ghg2i[g]:
                        ghg2i[g][hg] = len(ghg2i[g])

    clan_arrs = []
    for g, v in enumerate(groups):
        i0, i1 = v
        clan_arr = lil_matrix((i1 - i0, len(ghg2i[g])), dtype=bool)
#        clan_arr = np.zeros((i1-i0, len(ghg2i[g])), dtype=np.int8)
        for i2 in range(i0, i1):
            hgs = hg_loci[i2]
            for hg in hgs:
                if hg is not None:
#                    clan_arr[i2 - i0, ghg2i[g][hg]] = 1
                    clan_arr[i2 - i0, ghg2i[g][hg]] = True
        clan_arrs.append(clan_arr)


    blast_ids = defaultdict(dict)
    print('\t\t\t\tLoading alignment results', flush = True)
    if not os.path.isdir(f'{hlg_dir}algn/'):
        os.mkdir(f'{hlg_dir}algn/')
    jsons = collect_files(f'{hlg_dir}algn/', 'json')
    finished_hgs = set(int(os.path.basename(x).replace('.json', '')) \
                    for x in jsons)
    missing_hgs = set(blast_hash.keys()).difference(finished_hgs)
    if missing_hgs:
        blast_hash = {k: blast_hash[k] for k in list(missing_hgs)}
        with mp.get_context('forkserver').Pool(processes = cpus) as pool:
            pool.starmap(rnd2_gen_blastids_mp, 
                         ((hg, group_dict, hgx_dir, hlg_dir, minid) \
                           for hg, group_dict in blast_hash.items()))
            pool.close()
            pool.join()
    blast_ids = read_hlg_sim_jsons(f'{hlg_dir}algn/')

    # could be expedited by multiprocessing a writer and numba njit the math
    print('\t\t\t\tCalculating similarity', flush = True)
    if blast_ids:
        for g, v in enumerate(groups):
            i0, i1 = v
            clan_arr = clan_arrs[g]
            loc2loc_sim_mngr(clan_arr, finished_file, i0, loci, hg_loci,
                             0, blast_ids, min_hlg_id, simfun)
        # the file is now complete
        shutil.move(f'{hlg_dir}loci.adj.tmp.w', f'{hlg_dir}loci.adj.tmp')


def rnd1_loc2loc_mngr(
    db, clanI, loci, hgLoci, hg_dir, hgx_dir, hlg_dir, index,
    minid = 30, min_hlg_id = 0.15, simfun = overlap, min_overlap = 2,
    cpus = 1, Q = None
    ):

    finished_file = f'{hlg_dir}{clanI}.sadj.tmp'
    # is this clan already completed?
    if os.path.isfile(finished_file) or os.path.isfile(finished_file + '.r'):
        return
    # was this clan partially written?
    elif os.path.isfile(f'{finished_file}.w'):
        max_complete = open_prev_res(finished_file, index)
    # nothing completed
    else:
        max_complete = 0

    blast_hash = defaultdict(list)
    # only examine loci that have not been completed
    hg2i = {}
    for i in range(max_complete, len(loci)):
        locus = loci[i]
        hgs = hgLoci[i]
        for i1, hg in enumerate(hgs):
            if hg is not None:
                blast_hash[hg].append(locus[i1])
                if hg not in hg2i:
                    hg2i[hg] = len(hg2i)
    clan_arr = lil_matrix((len(loci) - max_complete, len(blast_hash)), dtype=bool)
    for i, v in enumerate(loci[max_complete:]):
        hgs = hgLoci[i + max_complete]
        for i1, hg in enumerate(hgs):
            if hg is not None:
                clan_arr[i, hg2i[hg]] = True


    blast_ids = defaultdict(dict)
    if cpus < 2:
        for hg, genes in blast_hash.items():
            gene_len = len(genes)
            if gene_len > 1:
                gene_set = set(genes)
                algn_base = f'{hgx_dir}{hg}.out'
                try:
                    blast_result = parse_blast(algn_base, gene_len, gene_set,
                                                minid)
                    if blast_result is not None:
                        blast_ids[hg] = blast_result                
                except FileNotFoundError:
                    pass

    else:
        print('\t\t\t\t\tLoading alignment results', flush = True)
        if not os.path.isdir(f'{hlg_dir}algn/'):
            os.mkdir(f'{hlg_dir}algn/')
        jsons = collect_files(f'{hlg_dir}algn/', 'json')
        finished_hgs = set(int(os.path.basename(x).replace('.json', '')) \
                        for x in jsons)
        missing_hgs = set(blast_hash.keys()).difference(finished_hgs)
        if missing_hgs:
            blast_hash = {k: blast_hash[k] for k in list(missing_hgs)}
            with mp.get_context('forkserver').Pool(processes = cpus) as pool:
                pool.starmap(gen_blastids_mp, 
                             ((hg, genes, hgx_dir, hlg_dir, minid) \
                               for hg, genes in blast_hash.items() \
                               if len(genes) > 1))
                pool.close()
                pool.join()
        blast_ids = read_hlg_sim_jsons(f'{hlg_dir}algn/')

    # could be expedited by multiprocessing a writer and numba njit the math
    if blast_ids:
        if cpus > 1:
            print('\t\t\t\t\tCalculating locus similarity', flush = True)
            Q.put(None)
        loc2loc_sim_mngr(clan_arr, finished_file, max_complete, loci, hgLoci,
                         index, blast_ids, min_hlg_id, simfun)

#        else:
 #           hlg_sim_mp_mngr(hits, f'{finished_file}.w', 
  #                          max_complete, loci, hgLoci, index, 
   #                         blast_ids, hlg_dir, f'{hlg_dir}mtrx/', min_hlg_id, 
    #                        simfun, chunkscale = 100, cpus = cpus)
        # begin appending to temporary writing output
        # the file is now complete
        shutil.move(f'{hlg_dir}{clanI}.sadj.tmp.w', f'{hlg_dir}{clanI}.sadj.tmp')
   #     if cpus > 1: # leave up to zipping 
#            jsons = collect_files(f'{hlg_dir}algn/', 'json')
 #           for j_f in jsons:
  #              os.remove(j_f)


def merge_clan_loci(preclan_loci, gene2hg, hgx2omes_path, hgx2dist_path, 
                    scaf2gene2i, merge_via_sim = False, 
                    min_branch_sim = 0.5, phylo = None):

    derep_loci = dereplicate_loci(preclan_loci, gene2hg, 
                                 hgx2omes_path, hgx2dist_path, scaf2gene2i,
                                 min_branch_sim, phylo)

    if merge_via_sim:
        vclan_loci = copy.deepcopy(derep_loci)
        out_clan_loci = defaultdict(lambda: defaultdict(list))
        outclanI = defaultdict(lambda: defaultdict(list))
        for clan, scaf_dict in derep_loci.items():
            for scaf, loci in scaf_dict.items():
                gene2i = scaf2gene2i[scaf]
                while True:
                    loc0, omes0 = loci[0]
                    try:
                        loc1, omes1 = loci[1]
                    except IndexError:
                        out_clan_loci[clan][scaf].append(loc0)
                        break
                    final_0 = gene2i[loc0[-1]]
                    start_1 = gene2i[loc1[0]]
                    if start_1 - final_0 == 1:
                        branch_sim = treecalcs.calc_branch_sim(phylo, omes0, omes1)
                        if branch_sim > min_branch_sim:
                            loci[1] = [loc0 + loc1, list(set(omes0).union(set(omes1)))]
                        else:
                            out_clan_loci[clan][scaf].append(loc0)
                    else:
                        out_clan_loci[clan][scaf].append(loc0)
                    loci.pop(0)
    else:
        out_clan_loci = {}
        for clan, scaf_dict in derep_loci.items():
           out_clan_loci[clan] = {}
           for scaf, loci in scaf_dict.items():
               out_clan_loci[clan][scaf] = [x[0] for x in loci]

    clan_loci = {clan: list(chain(*list(scaf_dict.values()))) \
                 for clan, scaf_dict in out_clan_loci.items()}

    hg_loci = {}
    for clan, loci in clan_loci.items():
        hg_loci[clan] = []
        for locus in loci:
            hg_loci[clan].append([])
            for gene in locus:
                try:
                    hg_loci[clan][-1].append(gene2hg[gene])
                except KeyError:
                    hg_loci[clan][-1].append(None)

    return clan_loci, hg_loci


def grab_contiguous_singletons(singletons, final_loci,
                               loc2hgx, gene2i,
                               hgx2omes, hgx2dist):

    window = max([len(v) for v in loc2hgx.values()])
    passing_genes = set(chain(*[x[0] for x in final_loci]))
    singletons_tmp = []
    for i, s in enumerate(singletons):
        if s[0] not in passing_genes:
            singletons_tmp.append(s)
    singletons = singletons_tmp

    final_singletons = []
    new_locs, new_dps = [], []
    contiguous = True
    count = 0
    while contiguous:
        count += 1
        singleton_i2g = {gene2i[v[0]]: v[0] for v in singletons}
        skip = set()
        singletons = sorted(singletons, key = lambda x: gene2i[x[0]])
        contiguous = set()
        for i, v in enumerate(singletons):
            s, null, omes0  = v
            if s in skip:
                continue
            skip.add(s)
            s_i = gene2i[s]
            b = s_i + 1
            contig = [s]
            while b in singleton_i2g:
                b_g = singleton_i2g[b]
                contig.append(b_g)
                b += 1 
                skip.add(b_g)
            if len(contig) < 3:
                final_singletons.append(v)
            else:
                contiguous.add(tuple(contig))
    
        new_singletons, todel, contiguous = [], [], list(contiguous)
        for ci, contig in enumerate(contiguous):
            t_new_locs = []
            for size in reversed(range(3, window + 1)):
                size_dec = len(contig) - size + 1
                if size_dec < 1:
                    continue
                for i, g in enumerate(contig[:size_dec]):
                    combo = tuple(contig[i:i+size])
                    if combo in loc2hgx:
                        hgx = loc2hgx[combo]
                        omes = hgx2omes[hgx]
                        dist = hgx2dist[hgx]
                        t_new_locs.append([list(combo), dist, omes])
            if not t_new_locs:
                todel.append(ci)
            t_new_locs = sorted(t_new_locs, key = lambda x: x[1], reverse = True)
            new_f_locs, new_singletons_p, new_dp = derep_loop(t_new_locs, gene2i)
            passing_genes = set(chain(*[x[0] for x in new_f_locs]))
            singletons_tmp = []
            for i, s in enumerate(new_singletons_p):
                if s[0] not in passing_genes:
                    singletons_tmp.append(s)
            new_singletons_p = singletons_tmp
        
            new_locs.extend(new_f_locs)
            new_dps.extend(new_dp)
            new_singletons.extend(new_singletons_p)
        for i in reversed(todel):
            del contiguous[i]
        singletons = new_singletons

    final_singletons.extend(singletons)

    return new_locs, new_dps
            

def derep_loop(loci, gene2i):
    final_loci, singletons_prep, dualtons = [], [], []
    while loci:
        locd0 = loci[0]
        loc0 = set(locd0[0])
        overi = 1
        for i1, locd1 in enumerate(loci[1:]):
            loc1 = set(locd1[0])

            locIntersection = loc0.intersection(loc1)
            if locIntersection:
    
                # index from the back because some may have hgx still
                dist1, omes1 = locd1[-2], locd1[-1]
                # the loci are sorted from largest to lowest distribution, so
                # we award the overlap to the largest TMD
                loc_dif = loc1.difference(loc0)
                # delete the initial version with overlap
                loci.pop(i1+overi)
                overi -= 1
                if loc_dif:
#                    if len(loc_dif) == 1:
 #                        singletons_prep.append([list(loc_dif)[0], dist1, omes1])
  #                       continue
                    new_locs = [sorted(loc_dif, key = lambda x: gene2i[x])]
                    break_i = True
                    # check for contuity, break-up noncontinuous loci
                    while new_locs:
                        new_loc = new_locs[0]
                        break_i = False
                        i_loc = [gene2i[x] for x in new_loc]
                        for i2, v in enumerate(i_loc[:-1]):
                            if not i_loc[i2+1] - v == 1:
                                break_i = i2 + 1
                                break
                        # if there's a break, reiterate and add singletons
                        if break_i:
                            n_loc0 = new_loc[:break_i]
                            n_loc1 = new_loc[break_i:]
              #              if len(n_loc0) > 1:
                            new_locs.append(n_loc0)
                #            else:
                 #               singletons_prep.append([n_loc0[0], dist1, omes1])
                  #          if len(n_loc1) > 1:
                            new_locs.append(n_loc1)
                            #else:
                             #   singletons_prep.append([n_loc1[0], dist1, omes1])
                        else:
                            loci.insert(i1+overi+1, [new_loc, dist1, omes1])
                            overi += 1
                        new_locs.pop(0)

        if len(loc0) > 2:
            final_loci.append(locd0)
        elif len(loc0) == 2:
            dualtons.append(locd0)
            for g in locd0[0]:
                singletons_prep.append([g, locd0[-2], locd0[-1]])
        else:
            singletons_prep.append([locd0[0][0], locd0[-2], locd0[-1]])
        loci.pop(0)
    return final_loci, singletons_prep, dualtons
        



def dereplicate_loci(preclan_loci, gene2hg,
                     hgx2omes_path, hgx2dist_path, scaf2gene2i,
                     min_branch_sim = 0.5, phylo = None):
    with open(hgx2dist_path, 'rb') as raw:
        hgx2dist = pickle.load(raw)
    with open(hgx2omes_path, 'rb') as raw:
        hgx2omes = pickle.load(raw)

    gene2hgx = defaultdict(list)
    out_clan_loci = defaultdict(lambda: defaultdict(list))
    for clan, scaf_dict in preclan_loci.items():
        for scaf, loci in scaf_dict.items():
            loc2hgx = {}
            gene2i = scaf2gene2i[scaf]
            for i, loc_d in enumerate(loci):
                loc, hgx = loc_d[0], loc_d[1]
                loc2hgx[tuple(loc)] = tuple(hgx)
                loci[i].append(hgx2dist[hgx])
                loci[i].append(hgx2omes[hgx])
 #               loci[i].append([])
                [gene2hgx[x].append(hgx) for x in loc]
            # sort the scaffolds loci from highest to lowest TMD
            loci = sorted(loci, key = lambda x: x[-2], reverse = True)
            # give the remaining overlapping genes to the highest distance group
            # will need to add the new HGxs in the end

            final_loci, singletons_prep, dualtons = derep_loop(loci, gene2i)


            add_loci, new_dp = grab_contiguous_singletons(singletons_prep, 
                                                  final_loci,
                                                  loc2hgx, gene2i,
                                                  hgx2omes, hgx2dist)  
            dualtons = sorted(dualtons, key = lambda x: x[-2], reverse = True)
            
            final_loci.extend(add_loci)

            genes_set = set(chain(*[x[0] for x in final_loci]))
            for d in dualtons:
                if all(x not in genes_set for x in d[0]):
                    final_loci.append(d)
            genes_set = set(chain(*[x[0] for x in final_loci]))
           
            gene2loc_i, i2gene, not_singleton = {}, {}, set()
            for i, locd in enumerate(final_loci):
                loc, omes = locd[0], locd[-1]
                for gene in loc:
                    gene2loc_i[gene] = [i, omes]
                    i2gene[gene2i[gene]] = gene
                    not_singleton.add(gene)

            singletons_prep = [x for x in singletons_prep \
                                if x[0] not in genes_set]
            # sort the singletons by TMD, highest to lowest to take the largest
            # branch length representative of the singleton
            sorted_s_prep = sorted(singletons_prep,
                                key = lambda y: y[-2], reverse = True)
            singletons, s_set = [], set()
            for s in sorted_s_prep:
                if s[0] not in s_set and s[0] not in not_singleton:
                    singletons.append(s)
                    s_set.add(s[0])


            singletons = sorted(singletons, key = lambda y: gene2i[y[0]])
            todel = True
            while todel:
                todel = []
                for i, v in enumerate(singletons):
                    s, null, omes0  = v
                    s_i = gene2i[s]
                    t, b = s_i - 1, s_i + 1
                    if t in i2gene:
                        t_g = i2gene[t]
                        t_i, t_o = gene2loc_i[t_g]
                        t_sim = treecalcs.calc_branch_sim(phylo, omes0, t_o)
                    else:
                        t_sim = 0
                    if b in i2gene:
                        b_g = i2gene[b]
                        b_i, b_o = gene2loc_i[b_g]
                        b_sim = treecalcs.calc_branch_sim(phylo, omes0, b_o)
                    else:
                        b_sim = 0
                    # if there are both directions, award to the most similar
                    # distribution
                    if t_sim > min_branch_sim and b_sim > min_branch_sim:
                        if t_sim == b_sim:
                            merge_base = random.choice([0, 1])
                            if merge_base:
                                merge_i, merge_o = t_i, t_o
                            else:
                                merge_i, merge_o = b_i, b_o
                        elif t_sim > b_sim:
                            merge_i = t_i
                            merge_o = t_o
                        else:
                            merge_i = b_i
                            merge_o = b_o
                    elif t_sim > min_branch_sim:
                        merge_i = t_i
                        merge_o = t_o
                    elif b_sim > min_branch_sim:
                        merge_i = b_i
                        merge_o = b_o
                    else:
#                        continuities_prep.append([s, null, omes0])
                        continue
                    gene2loc_i[s] = [merge_i, merge_o]
                    i2gene[gene2i[s]] = s
                    final_loci[merge_i][0].append(s)
                    todel.append(i)
                for i in reversed(todel):
                    singletons.pop(i)
                    
            continuities_iz = [gene2i[x[0]] for x in singletons]
            continuities = []
            for i, v in enumerate(continuities_iz[:-1]):
                if continuities_iz[i + 1] - v == 1:
                    continuities.append(i)

            merges = []
            for i in continuities:
                s0, s1 = singletons[i], singletons[i+1]
                omes0 = s0[-1]
                omes1 = s1[-1]
                branch_sim = treecalcs.calc_branch_sim(phylo, omes0, omes1)
                if branch_sim > min_branch_sim:
                    merges.append([{s0[0], s1[0]}, set(omes0).union(set(omes1))])

            for i0, locd0 in enumerate(merges):
                loc0, omes0 = locd0
                try:
                    loc1, omes1 = merges[i0+1]
                except IndexError:
                    if len(loc0) == 1:
                        merges[i0] = None
                    continue
                locIntersection = loc0.intersection(loc1)
                if locIntersection:
                    merges[i0 + 1] = [loc0.union(loc1), 
                                      omes0.union(omes1)]
                merges[i0] = None

            for i, locd in enumerate(merges):
                if locd:
                    loc, omes = locd
                    final_loci.append([list(loc), list(omes)])

            for locd in final_loci:
                loc, omes = locd[0], locd[-1]
                out_clan_loci[clan][scaf].append([sorted(loc, key = lambda x: gene2i[x]),
                                                  omes])


    clan_loci = {} 
    for clan, scaf_dict in out_clan_loci.items():
        clan_loci[clan] = {}
        for scaf, loci in scaf_dict.items():
            clan_loci[clan][scaf] = []
            for loc, omes in loci:
                clan_loci[clan][scaf].append([loc, omes])
            clan_loci[clan][scaf] = sorted(clan_loci[clan][scaf],
                                           key = lambda x: scaf2gene2i[scaf][x[0][0]])

  #  genes = Counter(list(chain(*list(chain(*[list(chain(*[y[0] for y in s_d.values()])) \
   #                 for s_d in out_clan_loci.values()])))))
    #if max(genes.values()) > 1:
     #   eprint('\t\t\tERROR: failed to dereplicate', flush = True)
      #  raise ValueError('ERROR: duplicate genes')


    return clan_loci #, out_hg_loci



def hash_clan_loci(ome, gff_path, ome_sig_clus, gene2hg, clusplusminus,
                   hgx2dist_path = None, hgx2omes_path = None,
                   merge_via_sim = False, min_branch_sim = 0.5, phylo = None):
    """sliding window of size clusplusminus, identify sequences that may have
    hgs that may be significant, grab their window, and check for a significant
    higher order hgx in the locus"""

    gff_list, protoclus, clus_out = gff2list(gff_path), {}, []
    cds_dict = input_parsing.compileCDS(gff_list, os.path.basename(gff_path).replace('.gff3',''))
    scaf2gene2i = {}
    for scaf, prots in cds_dict.items():
        scaf2gene2i[scaf] = {}
        for i, gene in enumerate(prots):
            scaf2gene2i[scaf][gene] = i

    preclan_loci = defaultdict(lambda: defaultdict(list))
    for scaf in cds_dict:
        for i0, seq0 in enumerate(cds_dict[scaf]):
            if seq0 in ome_sig_clus: # if the og is part of a significant seed
            # locus [[set(hgx)], i]
                for sig_clus in ome_sig_clus[seq0]:
                    clan = sig_clus[1]
                    start, end = None, None
                    if i0 < clusplusminus: # is i0 - clusplusminus < 0 ?
                        locus = cds_dict[scaf][:i0+clusplusminus+1] # then gather
                        # all the beginning
                    else:
                        locus = cds_dict[scaf][i0-clusplusminus:i0+clusplusminus+1]
                        # instead get the +/- and adjust for python
                    for i1, seq1 in enumerate(locus): # for each index and sequence
                    # in the locus
                        try:
                            hg = gene2hg[seq1]
                        except KeyError:
                            continue
                        if hg in sig_clus[0] and start is None: # if the og is
                        # in the sig clus and we haven't started
                            start = i1 # this is the start index
                        elif hg in sig_clus[0]: # otherwise if it is in the
                        # sig clus label this the other border unless another is
                        # found
                            end = i1 + 1
                    hgx = tuple(sorted(sig_clus[0]))
                    preclan_loci[clan][scaf].append([
                         locus[start:end], hgx
                         ])

    out_clan_loci, out_hg_loci = merge_clan_loci(preclan_loci, gene2hg,
                                                 hgx2omes_path,
                                                 hgx2dist_path,
                                                 scaf2gene2i, merge_via_sim,
                                                 min_branch_sim, phylo)
    return ome, out_clan_loci, out_hg_loci


def read_to_write(out_h, in_adj, min_hlg):
    with open(in_adj, 'r') as raw:
        for line in raw:
            d = line.rstrip().split()
            hlg_id = round(float(d[2]) * 100)
            if hlg_id >= min_hlg:
                out_h.write(line.rstrip() + '\n')
    out_h.flush()


def write_adj_matrix(out_file, hlg_dir, min_loc_id):
    comp_hlg_id = round(float(min_loc_id) * 100)
    with open(out_file + '.tmp', 'w') as out:
        # get previously read adjacency matrices
        init_files = collect_files(hlg_dir, 'sadj.tmp.r')
        for f in init_files:
            read_to_write(out, f, comp_hlg_id)

        # get ready to read adjacency matrices
        files = collect_files(hlg_dir, 'sadj.tmp')
        f_set = set(files)

        # while the completion signal has not been received
        while f'{hlg_dir}done.sadj.tmp' not in f_set:
            for f in files:
                read_to_write(out, f, comp_hlg_id)
                shutil.move(f, f'{f}.r')
            files = collect_files(hlg_dir, 'sadj.tmp')
            f_set = set(files)

        # get final files
        for f in files:
            if f == f'{hlg_dir}done.sadj.tmp':
                os.remove(f'{hlg_dir}done.sadj.tmp')
            else:
                read_to_write(out, f, comp_hlg_id)
                shutil.move(f, f'{f}.r')
    shutil.move(out_file + '.tmp', out_file)
    for f in collect_files(hlg_dir, 'sadj.tmp.r'):
        os.remove(f)


def check_tune(tune, hlg_hgxs, hlg_omes):
    failed = []
    satisfied = False
    for name, data in tune.items():
        hgx, omes, false_omes = data[0], data[1], data[2]
        set_hgx = set(hgx)
        set_omes = set(omes)
        set_false = set(false_omes)
        check = [v for i, v in enumerate(hlg_hgxs) \
                 if set_hgx.issubset(set(v)) \
                 and set_omes.issubset(hlg_omes[i]) \
                 and not set_false.intersection(hlg_omes[i])]
        if check:
            continue
        else:
            failed.append(name)
    if failed:
        for name in failed:
            print(f'\t\t\t{name} was not recovered', flush = True)
    else:
        print(f'\t\t\tTuned successfully', flush = True)
        satisfied = True
    return satisfied


def read_clus(clus_f, mci2pre):
    t_hlgs = defaultdict(list)
    with open(clus_f, 'r') as raw:
        for line in raw: # loci indices
            hlg, mci = [int(x) for x in line.rstrip().split()]
            t_hlgs[hlg].append(mci2pre[mci])

    return [v for k, v in sorted(t_hlgs.items(), key = lambda x: x[0]) \
            if len(v) > 1]
  #          indices = [int(x) for x in line.rstrip().split()]
            # singletons won't pass thresholds, disregard them
    #        if len(indices) > 1:
     #           t_hlgs.append(indices)



def mcl2hlgs(hlg_dir, loci, hgxXloci, ome2i, inflation, 
             loc2clan, tune = False, min_omes = 1, cpus = 1):
    satisfied = False
    if not os.path.isfile(hlg_dir + 'loci.clus') or tune:
        print(f'\t\t\tRunning MCL - inflation: {inflation}', flush = True)
        if cpus > 5:
            mcl_threads = 10 # cap at 5 processors, ten threads
        else:
            mcl_threads = (cpus * 2) - 1
        MCL(hlg_dir + 'loci.adj', hlg_dir + 'loci.clus',
                      inflation = inflation, threads = mcl_threads)
        # could add iterative subsample MCL option here

    mci2pre = read_mci_rows(hlg_dir + 'mcl_rows.tsv')
    t_hlgs = read_clus(hlg_dir + 'loci.clus', mci2pre)

    # list(hlg) = [{hgx: (omes,)}]
    hlgs, hlg_hgxs, hlg_omes, hlg2dom = [], [], [], {}
    with open(f'{hlg_dir}loci.txt', 'w') as out:
        out.write('#omeI locI clanI prehlg locus\n')
        for hlg, locIs in enumerate(t_hlgs):
#            hlgs.append(defaultdict(list))
            hlgs.append([])
            hlg_hgxs.append([])
            hlgOme_list = [] 
            locs = []
            for locI in locIs:
                loc = loci[locI]
                locs.append((loc, locI))
                hgxs = hgxXloci[locI]
                # really should be done above to make hgxs formatted right
                omeI = ome2i[loc[0][:loc[0].find('_')]]
#                [hlgs[-1][hgx].append(omeI) for hgx in hgxs]
                hlgs[-1].append((loc, locI))
                hlgOme_list.append(omeI)
                hlg_hgxs[-1].extend(hgxs)
            if len(set(hlgOme_list)) > min_omes: # need more than 1
#                hlgs[-1] = {k: sorted(set(v)) for k,v in hlgs[-1].items()}
                if hlgs[-1]:
                    hlgs[-1] = [tuple(x) for x in hlgs[-1]]
                    hlg_hgxs[-1] = tuple(sorted(set(hlg_hgxs[-1])))
                    hlg_omes.append(tuple(sorted(set(hlgOme_list))))
#                    hlg_hgxs.append(tuple(sorted(set(chain(*list(hlgs[-1].keys()))))))
  #                  hlg_omes.append(tuple(sorted(set(chain(*list(hlgs[-1].values()))))))
                    prehlg = len(hlgs) - 1
                    hlg2dom[prehlg] = loc2clan[locs[0][1]]
                    for i, omeI in enumerate(hlgOme_list):
                        locus, locI = locs[i]
                        out.write(f'{omeI} {locI} {loc2clan[locI]} {prehlg} {",".join(locus)}\n')
                else:
                    del hlgs[-1]
                    del hlg_hgxs[-1]
            else:
                del hlgs[-1]
                del hlg_hgxs[-1]

    failed = []
    if tune:
        satisfied = check_tune(tune, hlg_hgxs, hlg_omes)
    else:
        satisfied = True

    list_hlg2dom = []
    for i in range(len(hlgs)):
        list_hlg2dom.append(hlg2dom[i])

    return satisfied, {i: v for i, v in enumerate(hlgs)}, \
           {i: v for i, v in enumerate(hlg_hgxs)}, \
           {i: v for i, v in enumerate(hlg_omes)}, \
           {i: v for i, v in enumerate(list_hlg2dom)}

def parse_and_sort_gff(gff_path):
    # order the loci in each scaffold
    cds_dict = input_parsing.compileCDS(gff2list(gff_path),
                                        os.path.basename(gff_path).replace('.gff3',''))
    scaf2gene2i = {}
    for scaf, prots in cds_dict.items():
        scaf2gene2i[scaf] = {}
        for i, gene in enumerate(prots):
            scaf2gene2i[scaf][gene] = i
   
    return scaf2gene2i, cds_dict

def hash_ome_lg_loc(gff_path, locs_y_lg, max_between = 2):
    scaf2gene2i, cds_dict = parse_and_sort_gff(gff_path)
    scaf2i2gene = {}
    for scaf, gene2i in scaf2gene2i.items():
        scaf2i2gene[scaf] = {v: k for k, v in gene2i.items()}
    scaf2locs, null = output_data.clus2scaf(cds_dict, locs_y_lg)

    hlg2scaf2locs = defaultdict(lambda: defaultdict(list))
    for scaf, locs in scaf2locs.items():
        for loc, hlg in locs:
            hlg2scaf2locs[hlg][scaf].append(loc)
    
    hlg2locs = {}
    for hlg, scaf_dict in hlg2scaf2locs.items():
        hlg2locs[hlg] = []
        for scaf, locs in scaf_dict.items():
            while len(locs) > 1:
                loc0 = locs[0]
                loc1 = locs[1]
                top_i = scaf2gene2i[scaf][loc0[-1]]
                bot_i = scaf2gene2i[scaf][loc1[0]]
                if bot_i - top_i == 1:
                    locs[0].extend(locs[1])
                    locs.pop(1)
                # award multiple intervening genes if they meet the max_between criterium
                elif bot_i - top_i == max_between + 1:
                    locs[0].extend([scaf2i2gene[scaf][top_i + bw_i] \
                                    for bw_i in range(1, max_between + 1)] + locs[1])
                    locs.pop(1)
                else:
                    hlg2locs[hlg].append(loc0)
                    locs.pop(0)
            hlg2locs[hlg].append(locs[0])

    return hlg2locs


def lg_loc_mngr(db, lgs, gene2hg, max_between = 2, cpus = 1):
    omes2loci = defaultdict(list)
    for lg, locs in lgs.items():
        for loc, locI in locs:
            ome = loc[0][:loc[0].find('_')]
            omes2loci[ome].append((loc, lg))

    with mp.get_context('forkserver').Pool(processes = cpus) as pool:
        lg_loci = pool.starmap(hash_ome_lg_loc,
                               tqdm(((db[ome]['gff3'], omes2loci[ome],
                                      max_between) for ome in db.keys()), 
                                    total = len(db)))

    lg2locs = defaultdict(list)
    lg2hg_locs = defaultdict(list)
    lg2hgxs_locs = defaultdict(list)
    for res in lg_loci:
        for lg, locs in res.items():
            for loc in locs:
                hg_loc = []
                for gene in loc:
                    try:
                        hg_loc.append(gene2hg[gene])
                    except KeyError:
                        hg_loc.append(None)
                lg2locs[lg].append(loc)
                lg2hg_locs[lg].append(hg_loc)

    return lg2locs, lg2hg_locs    
#            db, clanI, locs, clan_hg_loci[clanI],
 #           hg_dir, hgx_dir, hlg_dir, index, minid, min_loc_id,
  #          simfun, min_overlap

def refine_group(db, ome2i, group_hg_loci, group_loci,
                 hg_dir, hgx_dir, wrk_dir, grp_dir, minid, min_loc_id,
                 simfun, inflation, tune, min_omes, prefix, cpus = 1, rnd = 1):

    m, adj_mtr = mp.Manager(), grp_dir + 'loci.adj'
    if cpus < 2:
        cpus = 2

#    min_overlap = round((min_loc_id * 2) - 0.5)
    min_overlap = 2 # need at least two genes of overlap to call it a cluster
    index, cmds = 0, []

    protog2h = {gI: Counter(chain(*hg_locs)) \
                for gI, hg_locs in enumerate(group_hg_loci)}
    group2hgx = {gI: set([k for k, v in h.items() if v > 1]) \
                 for gI, h in protog2h.items()}

    if rnd != 2:
        loci, hgxXloci, loc2group = {}, {}, {}
        for gI, locs in enumerate(group_loci):
            if not os.path.isfile(f'{grp_dir}{gI}.sadj.tmp.r'):
                cmds.append([
                    db, gI, locs, group_hg_loci[gI],
                    hg_dir, hgx_dir, grp_dir, index, minid, min_loc_id,
                    simfun, min_overlap
                    ])
            for i, locus in enumerate(locs):
                hg_loc = set(group_hg_loci[gI][i])
                loci[index] = group_loci[gI][i]
                hgxXloci[index] = tuple(sorted([x for x in \
                                       list(hg_loc.intersection(group2hgx[gI])) \
                                       if x is not None]))
                loc2group[index] = gI
                index += 1
        if not os.path.isfile(adj_mtr):
            W = mp.Process(target = write_adj_matrix, 
                           args = (adj_mtr, grp_dir, min_loc_id))
            W.start()
    #        print(f'\t\t\t{max([len(x) for x in )} loci in clan 0', flush = True)
            procs = []

            if cmds:
                if len(cmds[0][2]) > 30000 and cmds[0][1] == 0:
                    print(f'\t\t\t\tRunning group 0', flush = True)
                    Q = mp.Manager().Queue()
                    procs = [mp.Process(target=rnd1_loc2loc_mngr,
                                    args=(*cmds[0], cpus - 1, Q))]
                    procs[0].start()
                    Q.get()
                    print('\t\t\t\tRunning all groups', flush = True)
                    del cmds[0]
                with mp.get_context('forkserver').Pool(processes = cpus - 1) as pool:
                    pool.starmap(rnd1_loc2loc_mngr, tqdm(cmds, total = len(cmds), miniters = 1))
                    pool.close()
                    pool.join()
                if procs:
                    procs[0].join()
                    procs[0].close()
            with open(f'{grp_dir}done.sadj.tmp', 'w') as out:
                pass
            W.join()
    else:
        index, hgxXloci, loc2group, loci, hg_loci, groups = 0, {}, {}, {}, {}, []
        for gI, locs in enumerate(group_loci):
            groups.append([index, -1])
            for i, locus in enumerate(locs):
                hgs = group_hg_loci[gI][i]
                hg_loci[index] = hgs
                hg_loc = set(hgs)
                loci[index] = group_loci[gI][i]
                hgxXloci[index] = tuple(sorted([x for x in \
                                       list(hg_loc.intersection(group2hgx[gI])) \
                                       if x is not None]))
                loc2group[index] = gI
                index += 1
            groups[-1][1] = index
        if not os.path.isfile(f'{grp_dir}loci.adj.tmp') \
            and not os.path.isfile(f'{grp_dir}loci.adj'):
            rnd2_loc2loc_mngr(grp_dir, loci, hg_loci, hgx_dir, 
                              groups, minid, min_loc_id, simfun, cpus)
            os.rename(f'{grp_dir}loci.adj.tmp', f'{grp_dir}loci.adj')

    satisfied = False
    if tune:
        inflation = 1.1
    while not satisfied:
        if tune:
            inflation += 0.1
            if inflation > 3:
                break
        satisfied, grps, grp_hgxs, grp_omes, lilg2bigg = mcl2hlgs(grp_dir, loci,
                                                           hgxXloci, ome2i, 
                                                           inflation, loc2group,
                                                           tune, min_omes,
                                                           cpus)
    if tune:
        if not satisfied:
            eprint('\nERROR: tuning failed to recover clusters', flush = True)
            sys.exit(135)
        inf_strp = str(inflation * 10)[:2]
        inf_str = inf_strp[0] + '.' + inf_strp[1]
        with open(f'{wrk_dir}../log.txt', 'r') as raw:
            d = raw.read().replace(f'{prefix}_inflation\tNone', f'{prefix}_inflation\t' + inf_str)
        with open(f'{wrk_dir}../log.txt', 'w') as out:
            out.write(d)

    return grps, grp_hgxs, grp_omes, lilg2bigg, hgxXloci


def comp_scaf2hgs(cds_dict, gene2hg):
    scaf2hgs = {}
    for scaf, prots in cds_dict.items():
        scaf2hgs[scaf] = []
        for prot in prots:
            try:
                scaf2hgs[scaf].append(gene2hg[prot])
            except KeyError:
                scaf2hgs[scaf].append(None)
    return scaf2hgs


def extend_loci(ome, gff_path, loc2need, gene2hg, forbid_genes, 
                ext_len = 1, min_new_cont = 2, announcements = {87332}):
    forbidden_genes = set(forbid_genes)
    scaf2gene2i, cds_dict = parse_and_sort_gff(gff_path)
    
    gene2scaf, index_map = {}, {}
    for scaf, genes in cds_dict.items():
        index_map[scaf] = {v: i for i, v in enumerate(genes)}
        for gene in genes:
            gene2scaf[gene] = scaf

    # currently only built for an extension length of 1, will need modification
    # to accomodate greater lengths, e.g. determining how to merge an extension
    # if only the nearest-to-the-cluster entry of the extension meets missing_hg
    mod_locs, edge_locs = {}, {}
    for locI, d in loc2need.items():
        # opt to declare contig edges based on the starting locus, not the end
        # this is a decision that has arguments for/against both mechanisms
        # I'm setting it on the start to be conservative about this module
        contig_edge = False
        # keep track of modified loci, or contig edge loci for future mod
        modified = False
        loc, hlg, miss_hg_tup = d

        # announce this GCF for auditing
#        if gcf in announcements:
 #           announce = True
  #      else:
   #         announce = False

        miss_hg = set(miss_hg_tup)
        gene0 = loc[0]
        scaf = gene2scaf[gene0]
        scaf_len = len(cds_dict[scaf])
        ord_loc = sorted(loc, key = lambda pair: index_map[scaf][pair])
        # identify whether the top/bottom can be extended
        bot_i, top_i = scaf2gene2i[scaf][ord_loc[0]], scaf2gene2i[scaf][ord_loc[-1]]
        if bot_i:
            bot_i -= ext_len
            bot_ex = cds_dict[scaf][bot_i]
            # do NOT interfere with existing clusters
            if bot_ex in forbidden_genes:
                bot_ex = False
        else:
            contig_edge = True
            bot_ex = False
        if top_i < scaf_len - 1:
            top_i += ext_len
            top_ex = cds_dict[scaf][top_i]
            if top_ex in forbidden_genes:
                top_ex = False
        else:
            contig_edge = True
            top_ex = False

        # while there is a true extension possibility
        while top_ex or bot_ex:
            todel = set()
            if top_ex:
                try:
                    top_hg = gene2hg[top_ex]
                except KeyError:
                    top_hg = None
                if top_hg in miss_hg:
                    modified = True
                    ord_loc.append(top_ex)
                    todel.add(top_hg)
                    if top_i < scaf_len - 1:
                        top_i += ext_len
                        top_ex = cds_dict[scaf][top_i]
                        if top_ex in forbidden_genes:
                            top_ex = False
                else:
                    top_ex = False
            if bot_ex:
                try:
                    bot_hg = gene2hg[bot_ex]
                except KeyError:
                    bot_hg = None
                if bot_hg in miss_hg:
                    modified = True
                    ord_loc.insert(0, bot_ex)
                    todel.add(bot_hg)
                    if bot_i:
                        bot_i -= ext_len
                        bot_ex = cds_dict[scaf][bot_i]
                        if bot_ex in forbidden_genes:
                            bot_ex = False
                else:
                    bot_ex = False
            # dont edit the miss_hg until the end to not bias for/against top/bot    
            miss_hg = miss_hg.difference(todel)
        if modified:
            mod_locs[locI] = (ord_loc, hlg)
        
        # its risky to extend to different contigs based on a single HG, so we
        # should only do it if there's evidence that exceeds min_new_cont (2)
        if contig_edge and len(miss_hg) >= min_new_cont:
            edge_locs[locI] = (ord_loc, hlg, miss_hg)
        
    # if there are loci at the edge of a contig
    to_add_locs = []
    if edge_locs:
        scaf2hgs = comp_scaf2hgs(cds_dict, gene2hg)
        # we are only looking for extensions of min_new_cont length (2)
        # anything greater than this should have been picked up by a different
        # mechanism
        for locI, d in edge_locs.items():
            loc, hlg, miss_hg = d

            gene0 = loc[0]
            loc_scaf = gene2scaf[gene0]
            found = False
            # it will extend based on the first scaffold that meets the criteria
            for scaf, hg_list in scaf2hgs.items():
                # ignore self-scaffold check
                if loc_scaf == scaf:
                    continue
                bot_prot = cds_dict[scaf][:min_new_cont]
                top_prot = cds_dict[scaf][-min_new_cont:]
                # ignore other cluster entries
                if all(x not in forbidden_genes for x in bot_prot):
                    # this approach allows the rare situation for both sides of 
                    # the contig to meet the criteria before breaking
                    bot_set = hg_list[:min_new_cont]
                    if len(set(bot_set).intersection(miss_hg)) >= min_new_cont:
                        to_add_locs.append((bot_prot, hlg))
                        found = True
                else:
                    bot_set = set()
                if all(x not in forbidden_genes for x in top_prot):
                    top_set = hg_list[-min_new_cont:]
                    if len(set(top_set).intersection(miss_hg)) >= min_new_cont:
                        if top_set != bot_set:
                            to_add_locs.append((top_prot, hlg))
                            found = True
                if found:
                    break

    # ideally would like to propagate contig edge to downstream gbk because
    # antiSMASH does it and software like BiG-SCAPE use that data
    return ome, mod_locs, to_add_locs


def extend_loc_mngr(db, hlgs, hlg_hgxs, ome2gene2hg, hgxXloci, hlg_dir,
                    ext_len = 1, min_new_cont = 2, cpus = 1):
    ome2genes = defaultdict(list)
    ome2loc2need = defaultdict(dict)
    for hlg, locs in hlgs.items():
        hlg_hg_set = set(hlg_hgxs[hlg])
        for loc, locI in locs:
            ome = loc[0][:loc[0].find('_')]
            hg_loc = hgxXloci[locI]
            miss_hg = hlg_hg_set.difference(set(hg_loc))
            if miss_hg:
                ome2loc2need[ome][locI] = (loc, hlg, tuple(miss_hg))
            ome2genes[ome].extend(loc)
   
    with mp.Pool(processes = cpus) as pool:
        modifications = pool.starmap(extend_loci, 
                                     ((ome, db[ome]['gff3'], loc2need,
                                       ome2gene2hg[ome], ome2genes[ome],
                                       ext_len, min_new_cont) \
                                      for ome, loc2need in ome2loc2need.items()))

    with open(f'{hlg_dir}edge_loci.tsv', 'w') as edge_out:
        edge_out.write('#hlg\tlocus\n')
        with open(f'{hlg_dir}extension_loci.tsv', 'w') as extension_out:
            extension_out.write('#hlg\toriginal_locus\tupdate_locus\n')
            mod_count, cont_count = 0, 0
            for ome, mod_locs, to_add_locs in modifications:
                mod_count += len(mod_locs)
                for locI, d in mod_locs.items():
                    loc, hlg = d
                    ome2genes[ome].extend(loc)
                    for i, v in enumerate(hlgs[hlg]):
                        t_loc, t_locI = v
                        if t_locI == locI:
                            extension_out.write(f'{hlg}\t{",".join(t_loc)}\t{",".join(loc)}\n')
                            hlgs[hlg][i] = (loc, None)
                ome2genes[ome] = set(ome2genes[ome])
                for loc, hlg in to_add_locs:
                    # this awards the new loci to the first entry, which doesn't
                    # have an order that is necessarily meaningful
                    if all(x not in ome2genes[ome] for x in loc):
                        cont_count += 1
                        hlgs[hlg].append((loc, None))
                        edge_out.write(f'{hlg}\t{",".join(loc)}\n')
                        ome2genes[ome] = ome2genes[ome].union(set(loc))

        
    # we aren't modifying the hgs of each hlg so can ignore them
    return {hlg: tuple([loc[0] for loc in locs]) for hlg, locs in hlgs.items()}, \
           mod_count, cont_count


def classify_hlgs(
    hgx2loc, db, gene2hg, i2hgx, hgx2i,
    phylo, ome2i, hgx2omes, hg_dir, hgx_dir,
    wrk_dir, ome2partition, bord_scores_list,
    hg2gene, omes2dist = {}, clusplusminus = 3, inflation_1 = 1.1,
    inflation_2 = 1.5, min_hgx_overlap = 1, min_omes = 2, 
    minid = 30, cpus = 1, algorithm = 'diamond',
    min_loc_id = 0.3, simfun = overlap, printexit = False,
    tune = False, dist_func = treecalcs.calc_tmd,
    uniq_sp = False, min_branch_sim = 0.8, algn_sens = '',
    skipalgn = False, fallback = False, merge_via_sim = False,
    max_between = 2
    ):

    i2ome = {v: k for k, v in ome2i.items()}

    if not checkdir(hgx_dir, unzip = True, rm = True):
        os.mkdir(hgx_dir)

    groupI = wrk_dir + 'gene2hgx.pickle'
    groupII = wrk_dir + 'hgcs.pickle'

    # make an ome-centric hash for less MP overhead
    ome2gene2hg = defaultdict(dict)
    for gene, hg in gene2hg.items():
        ome = gene[:gene.find('_')]
        ome2gene2hg[ome][gene] = hg

    if not os.path.isfile(groupI):
        print('\tAggregating HGxs with overlapping loci', flush = True)
        hgx_genes = {}
        # compile a hash that shows the HGx each gene corresponds to
        for hgx in hgx2loc:
            if hgx in hgx2i: # if it passed the border threshold
                for gene in hgx2loc[hgx]:
                    ome = gene[:gene.find('_')]
                    if ome not in hgx_genes:
                        hgx_genes[ome] = {}
                    if gene not in hgx_genes[ome]:
                        hgx_genes[ome][gene] = []
                    hgx_genes[ome][gene].append(hgx)

        hashOgx_cmds = []
        start = datetime.now()
        for ome in hgx_genes:
            hashOgx_cmds.append([
                db[ome]['gff3'],
                ome, hgx_genes[ome], ome2gene2hg[ome], clusplusminus
                ])
        # expand the hgx 2 gene hash to include all genes in the locus corresponding
        # to the hgx
        with mp.get_context('fork').Pool(processes = cpus) as pool:
            gene_tups = pool.starmap(hash_hgx, hashOgx_cmds)
            pool.close()
            pool.join()    

        gene2hgx, hgx2genes = {}, {}

        for res in gene_tups:
            gene2hgx[ome2i[res[0]]] = {x[0]: set(x[1]) for x in res[1]}
            for hgx, hgs in res[2]:
                if hgx not in hgx2genes:
                    hgx2genes[hgx] = {og: [] for og in hgx}
                for og, seqs in hgs:
                    hgx2genes[hgx][og].extend(seqs)
            # hgx2genes = {hgx: og: [gene, gene]}
            # gene2hgx = {omeI: gene: {HGx}}

        with open(groupI, 'wb') as out:
            pickle.dump([gene2hgx, hgx2genes], out)
        print('\t\t' + str(datetime.now() - start))
    else:
        print('\tLoading aggregated HGxs', flush = True)
        with open(groupI, 'rb') as in_:
            gene2hgx, hgx2genes = pickle.load(in_)


    if not os.path.isfile(groupII):
        print(f'\tIdentifying HGxs with at least {min_hgx_overlap} gene overlap', 
              flush = True)
        tpairDict = defaultdict(list)
        for ome, gene2hgx_ome in tqdm(gene2hgx.items(),
                                      total = len(gene2hgx)):
            tpairDict = find_hgx_pairs(
                gene2hgx_ome, ome, hgx2i, tpairDict,
                min_hgx_overlap = min_hgx_overlap
                )

        pairDict = {}
        for id_, omes in tpairDict.items():
            omes_set = set(omes)
   #         if len(omes_set) > 1: # need at least one ome to calc a branch length
            pairDict[id_] = tuple(sorted(omes_set))
#        omes2dist = treecalcs.update_dists(phylo, pairDict, cpus = cpus, omes2dist = omes2dist,
 #                                          func = dist_func, uniq_sp = uniq_sp, i2ome = i2ome)


        print('\tAggregating HGxs', flush = True)
        # populate a lil_matrix here, then use that network to identify modules
        print('\t\tBuilding binary HGx-HGx network', flush = True)
        matrix = lil_matrix((len(i2hgx), len(i2hgx)), dtype=bool)
        for idPair, omes in tqdm(pairDict.items(), total = len(pairDict)):
            i0, i1 = idPair[0], idPair[1]
            matrix[i0, i1] = True
            matrix[i1, i0] = True
        print('\t\tIsolating subgraphs (HGx clans)', flush = True)
        g = Graph(directed = False)
        idx = matrix.nonzero()
        g.add_edge_list(np.transpose(matrix.nonzero()))
        components = label_components(g)[0]
        subgraph_is = sorted(set(components.a))
#        network = nx.from_scipy_sparse_matrix(matrix)
 #       subgraphs = [network.subgraph(c) for c in nx.connected_components(network)]
        # use .copy() after iterator to copy
        clans, clanOmes, clanHGxs = [], [], []
#        for sg in subgraphs:
        for sgi in subgraph_is:
            sg = GraphView(g, vfilt = components.a == sgi)
            clans.append({})
            clanOmes.append([])
            clanHGxs.append([])
#            for id_ in list(sg.nodes):
            for id_ in sg.vertices():
                hgx = i2hgx[int(id_)]
                omes = hgx2omes[hgx]
                clans[-1][hgx] = omes # storing in a hash to allow future
                # manipulation
                clanHGxs[-1].extend(hgx)
                clanOmes[-1].extend(omes)
            clanOmes[-1] = tuple(sorted(set(clanOmes[-1])))
            clanHGxs[-1] = tuple(sorted(set(clanHGxs[-1])))

        omes2dist = treecalcs.update_dists(
            phylo, {clanHGxs[i]: omes for i, omes in enumerate(clanOmes)},
            cpus = cpus, omes2dist = omes2dist, func = dist_func, uniq_sp = uniq_sp,
            i2ome = i2ome
            )
        with open(groupII, 'wb') as out:
            pickle.dump([clans, clanOmes, clanHGxs], out)
    else:
        print('\tLoading HGx clans', flush = True)
        with open(groupII, 'rb') as in_:
            clans, clanOmes, clanHGxs = pickle.load(in_)

    print('\t\t' + str(len(clans)) + ' clans', flush = True)
    print('\tClassifying HGxs into locus domains', flush = True)
    clanFile = wrk_dir + 'clan2loci.'
    if not os.path.isfile(clanFile + 'hg.json.gz'):
        print('\t\tPreparing locus extraction', flush = True)
        ome2clus2extract = defaultdict(lambda: defaultdict(list))
        for i, clan in enumerate(clans):
            for hgx, omes in clan.items():
                # create a seed hash of the hgx and its seed gene
                for gene in hgx2loc[hgx]:
                    ome = gene[:gene.find('_')]
                    ome2clus2extract[ome][gene].append([set(hgx), i])
                    # {ome: {gene: [[set(hgx), clanI]]}}

        print('\t\tExtracting clan loci', flush = True)
        with mp.get_context('forkserver').Pool(processes = cpus) as pool:
            hash_res = pool.starmap(hash_clan_loci, 
                                    tqdm(((ome, db[ome]['gff3'], clus2extract,
                                          ome2gene2hg[ome], clusplusminus,
                                          f'{wrk_dir}hgx2dist.pickle', 
                                          f'{wrk_dir}hgx2omes.pickle', merge_via_sim, min_branch_sim,
                                          phylo) \
                                          for ome, clus2extract \
                                          in ome2clus2extract.items()),
                                         total = len(ome2clus2extract)))
            pool.close()
            pool.join()

        print('\t\tWriting clan loci', flush = True)
        clan_loci = defaultdict(list)
#        clan_hgx4gcfs = defaultdict(list)
        clan_hg_loci = defaultdict(list)
        for ome, out_clan_loci, out_clan_hg_loci in hash_res:
            for clan in out_clan_loci:
                clan_loci[clan].extend(out_clan_loci[clan])
 #               clan_hgx4gcfs[clan].extend(out_clan_hgx[clan])
                clan_hg_loci[clan].extend(out_clan_hg_loci[clan])

        write_json(clan_loci, clanFile + 'json.gz')
  #      write_json(clan_hgx4gcfs, clanFile + 'hgx.json.gz')
        write_json(clan_hg_loci, clanFile + 'hg.json.gz')
    else:
        print('\t\tLoading clan loci', flush = True)
        clan_loci = {int(k): tuple(v) for k,v in sorted(
                read_json(clanFile + 'json.gz').items(), 
                key = lambda x: len(x[1]), reverse = True)}
   #     clan_hgx4gcfs = {int(k): tuple([tuple(i) for i in v]) \
    #                    for k,v in read_json(clanFile + 'hgx.json.gz').items()}
        clan_hg_loci = {int(k): tuple(v) \
                      for k, v in read_json(clanFile + 'hg.json.gz').items()}

    # dict(clan_loci) = {int(clanI): ((ome0_gene0, ome0_gene1,)...,)}
    # dict(clan_hgx4gcfs) = {int(clanI): ((og0, og1, ogn,)...,)}
    # dict(clan_hg_loci) = {int(clanI): ((ome0_gene0_og, ome0_gene1_og,)...,)}

    # need to move up to before writing
    clan_loci = {k: v for k,v in sorted(clan_loci.items(), key = lambda x: len(x[1]), reverse = True)}
    #clan_hgx4gcfs = [clan_hgx4gcfs[k] for k in clan_loci]
    clan_hg_loci = [clan_hg_loci[k] for k in clan_loci]
    clan_loci = [v for v in clan_loci.values()]

    hlg_dir = wrk_dir + 'hlg/' 
    if not os.path.isdir(hlg_dir):
        os.mkdir(hlg_dir)
    lg_dir = hlg_dir + 'lg/'
    if not os.path.isdir(lg_dir):
        os.mkdir(lg_dir)

    print('\t\tOutputting HG fastas', flush = True)
#    hgs_prep = set(chain(*list(chain(*list(clan_hg_loci.values())))))
    hgs_prep = list(chain(*list(chain(*clan_hg_loci))))
    hg_count = Counter(hgs_prep)
    if None in hg_count:
        del hg_count[None]
    hgs = sorted([x for x, v in hg_count.items() \
                  if v > 1])
    hg_dir = input_parsing.hg_fa_mngr(wrk_dir, None, 
                             hgs, db, hg2gene, cpus = cpus)
    evo_conco.run_blast(hgs, db, hg_dir, hgx_dir,
                        algorithm = algorithm, printexit = printexit,
                        sensitivity = algn_sens, hg2gene = hg2gene,
                        skipalgn = skipalgn, #minid = minid, let minid be at locus step
                        fallback = fallback, cpus = cpus)

    print('\t\tCalling locus domains (sub-cluster domains)', flush = True)
    clan_lens = [len(v) for v in clan_loci]
#    print('\t\t\t' + str(sum([len(v) for v in list(clan_loci.values())])) \
    print(f'\t\t\t{sum(clan_lens)}' \
        + ' loci', flush = True)
    print(f'\t\t\t{max(clan_lens)}' \
        + ' loci in clan 0', flush = True)
    lgs, lg_hgxs, lg_omes, lg2clan, hgxXloci = refine_group(
                 db, ome2i, clan_hg_loci, clan_loci,
                 hg_dir, hgx_dir, wrk_dir, lg_dir, minid, min_loc_id,
                 simfun, inflation_1, tune, min_omes, 'domain', cpus = cpus)
    print('\t\t\t' + str(len(lgs)) + ' domains', flush = True)

    print('\t\tMerging loci', flush = True)
    lg2locs, lg2hg_locs = lg_loc_mngr(db, lgs, gene2hg, max_between, cpus)
    lg_locs = {k: v for k,v in sorted(lg2locs.items(), key = lambda x: len(x[1]), reverse = True)}
    lg_hg_loci = [lg2hg_locs[k] for k in lg_locs]
    lg_locs = [v for v in lg_locs.values()]

    print('\t\tCalling HLGs (unfiltered-GCFs)', flush = True)
    lg_lens = [len(v) for v in lg_locs]
#    print('\t\t\t' + str(sum([len(v) for v in list(clan_loci.values())])) \
    print(f'\t\t\t{sum(lg_lens)}' \
        + ' loci', flush = True)
    print(f'\t\t\t{max(lg_lens)}' \
        + ' loci in LG 0', flush = True)
    hlgs, hlg_hgxs, hlg_omes, hlg2dom, hgxXloci = refine_group(
                 db, ome2i, lg_hg_loci, lg_locs,
                 hg_dir, hgx_dir, wrk_dir, hlg_dir, minid, min_loc_id,
                 sorensen, inflation_2, tune, min_omes, 'hlg', cpus = cpus, rnd = 2)
    print('\t\t\t' + str(len(hlgs)) + ' HLGs', flush = True)

    print('\tExtending incomplete loci and contig edge loci', flush = True)
    hlgs, mod_count, cont_count = extend_loc_mngr(db, hlgs, hlg_hgxs, ome2gene2hg, 
                                                  hgxXloci, hlg_dir, ext_len = 1,
                                                  min_new_cont = 2, cpus = cpus)
    print(f'\t\t{mod_count} loci extended', flush = True)
    print(f'\t\t{cont_count} loci merged from contig edges', flush = True)

    omes2dist = treecalcs.update_dists(
        phylo, {hlg_hgxs[i]: omes for i, omes in hlg_omes.items()},
        cpus = cpus, omes2dist = omes2dist, func = dist_func, 
        uniq_sp = uniq_sp, i2ome = i2ome
        )

    return hlgs, hlg_hgxs, hlg_omes, hlg2dom, omes2dist



import os
import re
import sys
import shutil
import pickle
import numpy as np
import multiprocessing as mp
from ast import literal_eval
from tqdm import tqdm
from datetime import datetime
from collections import defaultdict
from scipy import sparse
try:
    from numba import njit
    from numba.core.errors import NumbaDeprecationWarning, NumbaPendingDeprecationWarning
    @njit
    def gen_clusters_arr(loci, i2hg, clan_arr, check):
        """input is an HGp's 2D array of loci represented as a array of HGs
        presence absence.
        identify rows with > 2 overlapping HGs by multiplying the inputted
        2D np.array against its transposed self.
        if the overlapping loci are from different omes, report their
        overlapping HGs by pairwise loci summation and extracting coordinates
        with HG == 2 - indicating overlap.
        output these overlapping HGs (HGx), omes, and loci anchor genes""" 
        output = []
      #  adj_arr = clan_arr @ clan_arr.transpose() # sparse
        # e.g. hgs x loc     loc x hgs      loc x loc
        # [[1, 0, 1, 1],   [[1, 1, 1],    [[3, 3, 2],
        #  [1, 1, 1, 1], @  [0, 1, 1],  =  [3, 4, 3],
        #  [1, 1, 0, 1]]    [1, 1, 0],     [2, 3, 3]]
        #                   [1, 1, 1]]
        
        #  clan_arr         clan_arr.T      adj_arr
    
        # so in this example, loc 0 has an hgx (overlap > 2) with loc 0 and 1
        # loc 1 hgx with loc 0, 1, 2; and loc 2 an hgx with loc 1, 2
    
        # removing same locus and redundant overlap, there is thus a shared hgx
    
    
        adj_arr = clan_arr @ clan_arr.T # dense identify overlap
        for i0 in range(adj_arr.shape[0]): # for each row
            row = adj_arr[i0, :] # overlap row
            row0 = clan_arr[i0, :] # hg representation of locus
            loc0 = loci[i0] # the actual locus 
            ome0 = loc0[:loc0.find('_')] # the ome corresponding with the locus
    
            # these are loci coordinates non-self and non-redundant w/> 2 overlap
            overlap_loci = [x for x in np.where(row > 2)[0] if x > i0] # dense
    
            # for each of these overlaps
            for i1 in overlap_loci:
                loc1 = loci[i1]
                ome1 = loc1[:loc1.find('_')]
                if ome0 != ome1:
                    row1 = clan_arr[i1, :]
                    row_sum = row0 + row1
                    # if the sum of the row is 2, it indicates there is overlap
                    # at that particular HG
                    hgx = [i2hg[x] for x in np.where(row_sum >= 2)[0]] # dense
                    output.append((hgx, loc0, loc1))
        return output
except ImportError: # no numba 
    def gen_clusters_arr(loci, i2hg, clan_arr, check):
        """input is an HGp's 2D array of loci represented as a array of HGs
        presence absence.
        identify rows with > 2 overlapping HGs by multiplying the inputted
        2D np.array against its transposed self.
        if the overlapping loci are from different omes, report their
        overlapping HGs by pairwise loci summation and extracting coordinates
        with HG == 2 - indicating overlap.
        output these overlapping HGs (HGx), omes, and loci anchor genes""" 
        output = []
      #  adj_arr = clan_arr @ clan_arr.transpose() # sparse
        # e.g. hgs x loc     loc x hgs      loc x loc
        # [[1, 0, 1, 1],   [[1, 1, 1],    [[3, 3, 2],
        #  [1, 1, 1, 1], @  [0, 1, 1],  =  [3, 4, 3],
        #  [1, 1, 0, 1]]    [1, 1, 0],     [2, 3, 3]]
        #                   [1, 1, 1]]
        
        #  clan_arr         clan_arr.T      adj_arr
    
        # so in this example, loc 0 has an hgx (overlap > 2) with loc 0 and 1
        # loc 1 hgx with loc 0, 1, 2; and loc 2 an hgx with loc 1, 2
    
        # removing same locus and redundant overlap, there is thus a shared hgx
    
    
        adj_arr = clan_arr @ clan_arr.T # dense identify overlap
        for i0 in range(adj_arr.shape[0]): # for each row
            row = adj_arr[i0, :] # overlap row
            row0 = clan_arr[i0, :] # hg representation of locus
            loc0 = loci[i0] # the actual locus 
            ome0 = loc0[:loc0.find('_')] # the ome corresponding with the locus
    
            # these are loci coordinates non-self and non-redundant w/> 2 overlap
            overlap_loci = [x for x in np.where(row > 2)[0] if x > i0] # dense
    
            # for each of these overlaps
            for i1 in overlap_loci:
                loc1 = loci[i1]
                ome1 = loc1[:loc1.find('_')]
                if ome0 != ome1:
                    row1 = clan_arr[i1, :]
                    row_sum = row0 + row1
                    # if the sum of the row is 2, it indicates there is overlap
                    # at that particular HG
                    hgx = [i2hg[x] for x in np.where(row_sum >= 2)[0]] # dense
                    output.append((hgx, loc0, loc1))
        return output


import warnings

warnings.simplefilter('ignore', category=NumbaDeprecationWarning)
warnings.simplefilter('ignore', category=NumbaPendingDeprecationWarning)
from itertools import combinations, chain
from mycotools.lib.biotools import gff2list
from mycotools.lib.kontools import write_json, collect_files, read_json
from cloci.lib import treecalcs
from cloci.lib.input_parsing import compileCDS

def form_hgpDict(out_hgs):

    hgs = sorted(set(chain(*out_hgs)))
    hgp_dict = defaultdict(set)
    for hg0, hg1 in out_hgs:
        hgp_dict[hg0].add(hg1) # {og0: set(og1, ..., ogn), } or 
        # {og0: set(og0, og1, ..., ogn)}    
        hgp_dict[hg1].add(hg0)
                                            
    return hgp_dict


def hash_protohgxs(gff_path, hgp_dict, gene2hg, clusplusminus = 10):
    """parse a gff and compile its organized CDS_dict. identify where og-pairs
    co-occur and retrieve the set of hgs for the locus with the locus seed
    protein"""

    gff_list, protohgx = gff2list(gff_path), defaultdict(list)
    cds_dict = compileCDS(gff_list, os.path.basename(gff_path).replace('.gff3',''))

    for scaf in cds_dict:
        for i0, seq0 in enumerate(cds_dict[scaf]): # for each index and protein
            try:
                og0 = gene2hg[seq0]
            except KeyError: # gene not in OG / is a singleton
                continue
            if og0 in hgp_dict: # if this og is part of a significant seed
                if i0 < clusplusminus: # check if index - clusplusminus < 0
                    locus = cds_dict[scaf][:i0+clusplusminus+1] # obtain locus
                    # as all the sequences from the beginning of the contig to
                    # i0 plus the locus size tolerance (add 1 for python)
                else:
                    locus = cds_dict[scaf][i0-clusplusminus:i0+clusplusminus+1]
                    # otherwise obtain the locus from the +/-
                for seq1 in locus:
                    if seq1 != seq0: # if the protein isn't the first queried
                        try:
                            og1 = gene2hg[seq1]
                        except KeyError:
                            continue
                        if og1 in hgp_dict[og0]: # if it is in the set of
                        # hgs for the first queried og
                            og_loc_list = []
                            for gene in locus:
                                try:
                                    og_loc_list.append(gene2hg[gene])
                                except KeyError: # singleton or missing
                                    pass
                            og_loc = set(og_loc_list)
                            # the set of hgs in this locus
                            hgp = tuple(sorted([og0, og1])) # sort the hgs
                            protohgx[hgp].append([og_loc, seq0, seq1])
                            # {(og0, og1}: [[set(hgx, .. ogy), seq0, seq1], ...]}

    return dict(protohgx)


def merge_protos(protohgx_res):
    """combine protohgx results from each organism"""

    hgps2hgall = defaultdict(list)
    protohgx2omes = defaultdict(list)
    for ome_protohgx in protohgx_res: # for each protohgx from the mp
    # results
        for hgp, loc_d in ome_protohgx.items(): 
            protohgx2omes[hgp].extend(loc_d)
            for loc in loc_d:
                hgps2hgall[hgp].extend(list(loc[0]))
                    
            # {(hg0, hg1)}: {hg: i}}
    hgp2hgs = {hgp: {v: i \
                     for i, v in enumerate(sorted(set(hgs)))} \
               for hgp, hgs in hgps2hgall.items()}

    return {k: tuple(v) for k,v in protohgx2omes.items()}, hgp2hgs



def load_prehgx(wrk_dir, ome2i):
    h2l = defaultdict(list)
    pre_files = [os.path.basename(x) for x in collect_files(wrk_dir, 'txt')]
    preh_fs = [x for x in pre_files \
               if x.startswith('prehgx2data')]
    for preh_f in tqdm(preh_fs, total = len(preh_fs)):
        with open(f'{wrk_dir}{preh_f}', 'r') as raw:
            for line in raw:
                d = line.rstrip().split()
                hgx = tuple((int(x) for x in d[0].split(',')))
                gs = d[1].split(',')
                h2l[hgx].extend(gs)

    h2l = {k: tuple(sorted(set(v))) for k, v in h2l.items()}
    h2o = {k: tuple(sorted(set(ome2i[y[:y.find('_')]] for y in v))) \
           for k, v in h2l.items()}

    return h2o, h2l


def hgx2mngr(Q, wrk_dir, suffix):
    """Receive from gen_clusters and formulate_clus_dict shared Manager Queue
    Output the data to drive storage for parsing and collection later"""
    x = True
    hgp_file = f'{wrk_dir}prehgx.{suffix}.txt'
    dat_file = f'{wrk_dir}prehgx2data.{suffix}.txt'
    with open(hgp_file, 'a') as hgp_out, open(dat_file, 'a') as dat_out:
        while x:
            x = Q.get()
            if x:
                for mhgx, g0, g1 in x[1]:
                    out_str = ','.join([str(x) for x in mhgx])
                    out_str += f' {g0},{g1}\n'
                    dat_out.write(out_str)
    
                hgp = x[0]
                hgp_out.write(f'{hgp[0]} {hgp[1]}' + '\n')
                hgp_out.flush()
                dat_out.flush()
    Q.put(None)


def formulate_hgx_dicts(hgp, loci, hgs2i, Q):
    hgp_set = set(hgp)
    # make a matrix to identify overlap hgs v loci
    # loci must a be sorted by hgs
    clan_arr = np.zeros([len(loci), len(hgs2i)],
                        dtype = np.int8)
    for i, loc in enumerate(loci):
        hg_loc = np.array(list([hgs2i[x] for x in loc[0]]))
        clan_arr[i, hg_loc] = 1
    loci = [x[1] for x in loci]

#    if len(set(hgs2i.keys()).intersection({10, 1025, 6380, 10267, 12567, 63630})) > 2:
 #       print(loci, flush = True)
    check = False
    if hgp in {(10, 10267), (10, 12567), (10267, 12567), (1025, 63630), (10, 63630)}:
        check = True
    i2hg = np.array([k for k, v in hgs2i.items()])
    output = gen_clusters_arr(loci, i2hg, clan_arr.astype(float), check)
#                output.append([hgx, (ome0, ome1), (loc0, loc1)])
    Q.put((hgp, tuple((tuple(x[0]), x[1], x[2],) \
           for x in output),))


def par_rm(hgx, hgx2omes):

#    check_set, out_p = {10, 1025, 6380, 10267, 12567, 63630}, False
    t_comb_sets, t_combs, todel = [set(hgx)], [hgx], []
    # for each length of combinations possible
    for comb_len in reversed(range(3, len(hgx))):
        # for each combination at that length
        for comb in combinations(hgx, comb_len):
            # if this combination already exists 
            if comb in hgx2omes:
                # get the set of this subhgx and check if its a subset
                set_comb = set(comb)
                for i, set_hgx in enumerate(t_comb_sets):
                    # if it is a subset
                    if set_comb < set_hgx:
                        # and if it has the same omes as the superset
                        if hgx2omes[comb] == hgx2omes[t_combs[i]]:
                            todel.append(comb)
                            break
                    t_combs.append(comb)
                    t_combs_sets.append(set_comb)

    return todel


# I think most subsets will be accounted from the initial comparison,
# and the ones that aren't weren't part of some significant HGp.... soo
# they weren't significant - this would be the most cautious approach,
# but a lot of overhead for something

# in other words, this would only discover few HGxs that are part of
# loci with an HGp that is not part of the HGx; so it recoups some
# of the significant HGxs overlooked from the HGp seeding algorithm,
# but at the end of the day HGxs that are part of nonsignificant seed
# loci are still overlooked. 

# the subset HGxs that do exist will be populated because the loci will
# be overlapped - the performance hit doesnt justify this. if you want
# these loci then decrease the HGp percentile
def add_subsets(hgx2omes, hgx2loc):
    for mhgx in list(hgx2omes.keys()):
        omes = hgx2omes[mhgx]
        for c_len in range(3, len(mhgx)):
            for hgx in combinations(mhgx, c_len):
                hgx2omes[hgx].update(omes)
                hgx2loc[hgx].update(hgx2loc[hgx])

    return hgx2omes, hgx2loc

def rm_subsets(hgx2omes, hgx2loc, cpus = 1):

#    hgx2omes, hgx2loc = add_subsets(hgx2omes, hgx2loc)
    max_len = max((len(x) for x in list(hgx2omes.keys())))
    for hgx_len in tqdm(reversed(range(4, max_len + 1)),
                        total = max_len + 1 - 4):
        i2hgx = list(hgx2omes.keys())
        rm_cmds = [[hgx, hgx2omes] for hgx in i2hgx if len(hgx) == hgx_len]
        with mp.get_context('fork').Pool(processes = cpus) as pool:
            todels = pool.starmap(par_rm, rm_cmds)
            pool.close()
            pool.join()
        
        for todel in todels:
            for hgx in todel:
                try:
                    del hgx2omes[hgx]
                    del hgx2loc[hgx]
                except KeyError:
                    continue

    return hgx2omes, hgx2loc


def id_hgx(db, hgp_dict, gene2hg, ome2i, wrk_dir, cpus, clusplusminus = 10):
    """form hgxs from hgps by extracting higher order combinations from
    overlapping hgx loci"""

    hash_protohgx_cmds = []
    gffs = [v['gff3'] for k, v in db.items() if k in ome2i]
    print('\t\tForming proto-HGx hashes', flush = True)
    for gff in gffs: # prepare for protohgx hashing by organism
        hash_protohgx_cmds.append((gff, hgp_dict, gene2hg, clusplusminus,))
    with mp.get_context('fork').Pool(processes = cpus) as pool:
        protohgx_res = pool.starmap(hash_protohgxs, 
                                    tqdm(hash_protohgx_cmds,
                                         total = len(hash_protohgx_cmds)))
        pool.close()
        pool.join()
                
    print('\t\tForming proto-HGxs', flush = True)
    protohgx2omes, hgp2hgs = merge_protos(protohgx_res) # merge mp results
            
    print('\t\tIdentifying HGxs', flush = True)
    count, ite, proc = 0, 1, None
    
    pre_files = [os.path.basename(x) for x in collect_files(wrk_dir, 'txt')]
    if any(x.startswith('prehgx') for x in pre_files):
        print('\t\t\tLoading checkpoint', flush = True)
        preh_fs = [x for x in pre_files if x.startswith('prehgx.')]
        sxs = [int(re.search(r'prehgx\.(\d+)\.txt', x)[1]) for x in preh_fs]
        sx = max(sxs)
        hgps = []
        for preh_f in tqdm(preh_fs, total = len(preh_fs)):
            with open(f'{wrk_dir}{preh_f}', 'r') as raw:
                for line in raw:
                    d = line.rstrip().split()
                    if len(d) > 1:
                        hgps.append((int(d[0]), int(d[1]),))
        hgps = list(set(hgps))
        for hgp in hgps:
            if hgp in protohgx2omes:
                del protohgx2omes[hgp]
    else:
        sx = -1

    if protohgx2omes:
#        prog = mp.Process(target=progress_report, args=(wrk_dir, 0, start_len))
 #       prog.start()
        # could make this intelligent via a test
        if cpus == 1:
            eprint('ERROR: need more than one CPU', flush = True)
        if cpus == 2:
            c_cpus = 1
            w_cpus = 1
        elif cpus < 3:
            w_cpus = 1
            c_cpus = cpus - 1
        else:
            w_cpus = round(cpus/3)
            c_cpus = cpus - w_cpus
    
        Q = mp.Manager().Queue()
        procs = []
        for i in range(w_cpus):
            sx += 1
            procs.append(mp.Process(target=hgx2mngr, args=(Q, wrk_dir, sx)))
            procs[-1].start()

        params = ((hgp, loci, hgp2hgs[hgp], Q) for hgp, loci in protohgx2omes.items()) 
        with mp.Pool(processes = c_cpus) as pool:
            pool.starmap(formulate_hgx_dicts, tqdm(params, total = len(protohgx2omes)))
            pool.close()
            pool.join()
                
        Q.put(None)
        for proc in procs:
            proc.join()
        Q.get()

   # print('joining', flush = True)
  #  prog.join()
    print('\t\tReading results', flush = True)
    hgx2omes, hgx2loc = load_prehgx(wrk_dir, ome2i)

 #   print('\t\tRemoving subset HGxs', flush = True)
#    hgx2omes, hgx2loc = rm_subsets(hgx2omes, hgx2loc, cpus)
 #   hgx2omes = {x: tuple(sorted(hgx2omes[x])) for x in hgx2omes}

    return hgx2omes, hgx2loc


def hgp2hgx(db, wrk_dir, top_hgs, gene2hg, ome2i, 
                phylo, plusminus = 3, cpus = 1):
    if not os.path.isfile(wrk_dir + 'hgx2loc.pickle'):
        hgp_dict = form_hgpDict(top_hgs)
        print('\tForming HGxs', flush = True)
        form_clus_start = datetime.now()
        hgx2omes, hgx2loc = id_hgx(
            db, hgp_dict, gene2hg,  
            ome2i, wrk_dir, cpus, clusplusminus = plusminus
            )
#        write_json({str(k): list(v) for k, v in hgx2loc.items()},
 #                  f'{wrk_dir}hgx2loc.json.tmp.gz')
        with open(wrk_dir + 'hgx2loc.pickle', 'wb') as pickout:
            pickle.dump(hgx2loc, pickout)
 #       write_json({str(k): list(v) for k, v in hgx2omes.items()},
  #                 f'{wrk_dir}hgx2omes.json.tmp.gz')
   #     shutil.move(f'{wrk_dir}hgx2loc.json.tmp.gz', 
    #                f'{wrk_dir}hgx2loc.json.gz')
     #   shutil.move(f'{wrk_dir}hgx2omes.json.tmp.gz',
      #              f'{wrk_dir}hgx2omes.json.gz')
        with open(wrk_dir + 'hgx2omes.pickle', 'wb') as pickout:
            pickle.dump(hgx2omes, pickout)
        formHGxTime = datetime.now() - form_clus_start
        pre_files = [os.path.basename(x) for x in collect_files(wrk_dir, 'txt')]
        todel = [x for x in pre_files if x.startswith('prehgx')]
        for f in todel:
            os.remove(wrk_dir + f)
        print('\t\t' + str(formHGxTime), flush = True)
    
    else: # or just load available structures 
#        hgx2loc = literal_load(f'{wrk_dir}hgx2loc.json.gz')
 #       hgx2omes = literal_load(f'{wrk_dir}hgx2omes.json.gz')
        with open(wrk_dir + 'hgx2loc.pickle', 'rb') as pickin:
            hgx2loc = pickle.load(pickin)
        hgx2omes = {k: tuple(sorted(set(ome2i[y[:y.find('_')]] for y in v))) \
                    for k, v in hgx2loc.items()}
#        with open(wrk_dir + 'hgx2omes.pickle', 'rb') as pickin:
 #           hgx2omes = pickle.load(pickin)

    return hgx2omes, hgx2loc



#! /usr/bin/env python3

import os
import sys
import gzip
import shutil
import argparse
import multiprocessing as mp
from tqdm import tqdm
from itertools import chain
from cloci.lib.output_data import ann_mngr, annotate_clusters
from mycotools.db2search import compile_hmm_queries
from mycotools.lib.dbtools import mtdb, primaryDB
from mycotools.lib.kontools import format_path, eprint, mkOutput, findExecs, \
    collect_files

# NEED variable hmmsearch args
# NEED to output annotations for hlg runs and gcfs runs and reference HLGs from gcf if present

def input_annotations(ann_file):
    # 0 = no change, 1 = and condition, 2 = or condition
    ann_change = False
    acc2ann = {}
    with open(ann_file, 'r') as raw:
        for line in raw:
           if not line.startswith('#'):
               d = line.rstrip().split('\t')
               if len(d) > 1:
                   ann_change = True
                   if ';' not in d[0] and '|' not in d[0]:
                       acc2ann[d[0]] = [None, d[1]]
                   else:
                       ands = d[0].split(';')
                       and_ors = [x.split('|') for x in ands]
                       for a in and_ors:
                           for o in a:
                               if ';' in o or '|' in o:
                                   eprint('\nERROR: multi-level nested annotations ' \
                                        + 'not supported', flush = True)
                                   sys.exit(5)
                               acc2ann[o] = [[set(x) for x in and_ors if x != a],
                                             d[1]]
                  
               elif len(d) == 1:
                   acc2ann[d[0]] = [None, d[0]]
    return acc2ann, ann_change


def apply_annotations(hlg_file, out_file, acc2ann, extract = False):
    data, count, passing, gcfs = [], 0, set(), []
    with open(hlg_file, 'r') as raw:
        for line in raw:
            clus_type = set()
            d = line.rstrip().split('\t')
            if line.startswith('#'):
                data.append(d)
                continue
            if len(d) == 6: # previous annotations
                prev_anns = [x.split('|') if x else [] \
                             for x in d[-2].split(';')]
            else:
                prev_anns = None
            anns = [x.split('|') if x else [] for x in d[-1].split(';')]
            a_set = set(chain(*anns))
            for i0, v in enumerate(anns):
                for i1, ann in enumerate(v):
                    if ann in acc2ann:
                        if acc2ann[ann][0] and acc2ann[ann][1] not in clus_type:
                            trues = 0
                            ands_len = len(acc2ann[ann][0])
                            for etc in acc2ann[ann][0]:
                                if any(x in a_set for x in etc):
                                    trues += 1
                            if trues == ands_len:
                                clus_type.add(acc2ann[ann][1])       
                        elif not acc2ann[ann][0]:
                            clus_type.add(acc2ann[ann][1]) 
                        anns[i0][i1] = acc2ann[ann][1]
            if any(x for x in clus_type):
                d[0] += '_' + '$'.join(sorted(clus_type))
                if '' in clus_type:
                    clus_type.remove('')
                passing.add(count)
                gcfs.append(int(d[3]))
            if prev_anns:
                for i, v in enumerate(anns):
                    if not v:
                        if prev_anns[i]:
                            anns[i] = prev_anns[i]
                d[-1] = ';'.join(['|'.join([x for x in y]) for y in anns])
                del d[-2]
            elif not prev_anns:
                d[-1] = ';'.join(['|'.join([x for x in y]) for y in anns])
            data.append(d)
            count += 1

    with open(out_file, 'w') as out:
        out.write('\t'.join(data[0]) + '\n')
        if extract:
            for i, v in enumerate(data[1:]):
                if i in passing:
                    out.write('\t'.join(v) + '\n')
        else:
            for d in data[1:]:
                out.write('\t'.join(v) + '\n')

    return sorted(set(gcfs))

def import_hlgs(ome, hlg_file):
    gene2hg, gene2ann, clus2gene, clus2gcf = {}, {}, {}, {}
    with open(hlg_file, 'r') as raw:
        for line in raw:
            if not line.startswith('#'):
                d = line.rstrip().split('\t')
                h, g = d[1].split(','), d[2].split(',')
                for i, gene in enumerate(g):
                    gene2hg[gene] = h[i]
    return ome, gene2hg


def main(hmm_paths, cloci_dir, in_dir, out_dir, db, prefix = 'hlg', 
         extract = False, ann_file = None, cpus = 1):

    db = db.set_index()

    print('\nCompiling input', flush = True)
    wrk_dir = out_dir + 'working/'
    if not os.path.isdir(wrk_dir):
        os.mkdir(wrk_dir)
    ome_dir = out_dir + 'ome/'
    if not os.path.isdir(ome_dir):
        os.mkdir(ome_dir)

    if ann_file:
        acc2ann, ann_change = input_annotations(ann_file)
    else:
        acc2ann, ann_change = {}, False


    q_hmm = wrk_dir + 'query.hmm'
    queries = compile_hmm_queries(hmm_paths, q_hmm)


    extract_cmds = []
    for hlg_file in collect_files(in_dir, 'tsv', recursive = True):
        if os.path.basename(hlg_file) == f'{prefix}.tsv':
            ome = os.path.basename(os.path.dirname(hlg_file))
            if ome not in db:
                continue
            if not os.path.isdir(ome_dir + ome):
                os.mkdir(ome_dir + ome)
            extract_cmds.append((ome, hlg_file))
            shutil.copy(hlg_file, ome_dir + ome + '/old.tsv')
            
    with mp.Pool(processes = cpus) as pool:
        out_res = pool.starmap(import_hlgs, tqdm(extract_cmds, total = len(extract_cmds)))

    ome2gene2hg = {}
    prot_paths = {}
    for ome, gene2hg in out_res:
        ome2gene2hg[ome] = gene2hg
        prot_paths[ome] = db[ome]['faa']

    genes_dict = {ome: [list(genes.keys())] for ome, genes in ome2gene2hg.items()}
    if not genes_dict:
        print('\nNothing to do', flush = True)
        sys.exit(0)

    print('\nAnnotating', flush = True)
    ann_res, failed_omes = ann_mngr(genes_dict, prot_paths, wrk_dir, q_hmm, False,
                                    evalue = 0.01,
                                    threshold = 0.35, cpus = cpus)
    eprint(f'\t\tWARNING: Failures: {",".join(sorted(failed_omes))}', flush = True)
    for ome in failed_omes:
        del ome2gene2hg[ome]


    print('\nApplying annotations', flush = True)
    annotation_cmds = []
    for ome, gene2hg in ome2gene2hg.items():
        annotation_cmds.append((list(gene2hg.keys()), ome, #f'{ome_dir}{ome}/old.tsv',
                                ome_dir, #f'{ome_dir}{ome}/{os.path.basename(hlg_file)}',
                                gene2hg, 'old', ann_res[ome]))

    with mp.Pool(processes = cpus) as pool:
        pool.starmap(annotate_clusters, tqdm(annotation_cmds, 
                                             total = len(annotation_cmds)))

    if extract or ann_change:
        print('\nConverting annotations', flush = True)
        with mp.Pool(processes = cpus) as pool:
            gcfs_res = pool.starmap(apply_annotations, 
                         tqdm(((f'{ome_dir}{ome}/old.tsv',
                               f'{ome_dir}{ome}/{prefix}.tsv',
                               acc2ann, extract) for ome in ome2gene2hg),
                              total = len(ome2gene2hg)))
        gcfs = set(chain(*sorted(gcfs_res)))
        out = ''
        with gzip.open(f'{cloci_dir}hlgs.tsv.gz', 'rt') as raw:
            for line in raw:
                if line.startswith('#'):
                    out += line
                else:
                    d = line.rstrip().split()
                    t_gcf = int(d[1])
                    if t_gcf in gcfs:
                        out += line
        with gzip.open(f'{out_dir}{prefix}s.tsv.gz', 'wt') as out_f:
            out_f.write(out)
    else:
        shutil.copy(f'{cloci_dir}hlgs.tsv.gz', 
                   f'{out_dir}hlgs.tsv.gz')
        

def cli():
    parser = argparse.ArgumentParser(
        description = 'Annotates HLGs/GCFs. Will overwrite gcfs if -e is specified and not -o')
    parser.add_argument('-i', '--input', help = 'HMM dir/file', required = True)
    parser.add_argument('-c', '--cloci', help = 'CLOCI|CLOCI/ome dir', required = True)
    parser.add_argument('-d', '--mtdb', default = primaryDB())
    parser.add_argument('-a', '--annotations', 
        help = 'File with annotation conversions: <ACC>\\t<ANN>; `;` = AND `|` = OR')
    parser.add_argument('-g', '--gcf', help = 'Extract from GCFs', action = 'store_true')
    parser.add_argument('-e', '--extract', help = 'Extract annotations', action = 'store_true')
    parser.add_argument('--cpus', type = int, default = 1)
    args = parser.parse_args()

    findExecs(['hmmsearch'], exit = ['hmmsearch'])

#    if os.path.basename(format_path(args.cloci)[:-1]) == 'ome':
 #       out_dir = mkOutput(format_path('./'), 'cloci2annotation') 
  #      ome_dir = format_path(args.cloci)
   # else:
    out_dir = mkOutput(format_path(args.cloci), 'cloci2annotation', suffix = False)
    ome_dir = format_path(args.cloci) + 'ome/'

    if os.path.isdir(format_path(args.input)):
        hmm_files = collect_files(format_path(args.input), 'hmm')
    elif os.path.isfile(format_path(args.input)):
        hmm_files = [format_path(args.input)]
    else:
        eprint('\nERROR: invalid -i')
        sys.exit(3)

    if args.gcf:
        prefix = 'gcf'
    else:
        prefix = 'hlg'    

    main(hmm_files, format_path(args.cloci), ome_dir, out_dir, 
         mtdb(format_path(args.mtdb)), prefix, 
         extract = args.extract, ann_file = format_path(args.annotations), 
         cpus = args.cpus)


if __name__ == '__main__':
    cli()
    sys.exit(0)



#! /usr/bin/env python3

import os
import sys
import gzip
import argparse
import multiprocessing as mp
from tqdm import tqdm
from itertools import chain
from collections import defaultdict
from mycotools.acc2gbk import ome_main as acc2gbk
from mycotools.acc2gff import gff_main as acc2gff
from mycotools.lib.dbtools import primaryDB, mtdb
from mycotools.lib.biotools import gff2list, fa2dict, list2gff, dict2fa
from mycotools.lib.kontools import stdin2str, format_path, eprint


def parse_hlgtsv(hlg_tsv):
    hlg2omes = {}
    if hlg_tsv.endswith('.gz'):
        with gzip.open(hlg_tsv, 'rt') as raw:
            for line in raw:
                if not line.startswith('#'):
                    d = line.rstrip().split()
                    hgs_s, hlg_s, hlc_s, tmd_s, p_s, g_s, mi_s, mp_s, ippr_s, o_s = d
                    hlg, omes = int(hlg_s), o_s.split(',')
                    hlg2omes[hlg] = omes

    omes2hlg = defaultdict(list)
    for hlg, omes in hlg2omes.items():
        for ome in omes:
            omes2hlg[ome].append(hlg)

    return hlg2omes, omes2hlg
                     

def parse_ome(ome, ome_file, hlgs = []):
    hlg_set = set(hlgs)

    hlg2genes = defaultdict(list)
    with open(ome_file, 'r') as raw:
        if not hlg_set:
            for line in raw:
                if not line.startswith('#'):
                    d = line.rstrip().split()
                    genes, hlg = d[2].split(','), int(d[3])
                    hlg2genes[hlg].append(genes)
        else:
            for line in raw:
                if not line.startswith('#'):
                    d = line.rstrip().split()
                    genes, hlg = d[2].split(','), int(d[3])
                    if hlg in hlg_set:
                        hlg2genes[hlg].append(genes)
    return ome, hlg2genes

        
def compile_ome_gffs(row, ome, hlg2genes, out_dir, by_ome = False, gbks = False, gffs = False):
    gff = gff2list(row['gff3'])
    hlg2gffs = defaultdict(list)
    for hlg, gene_ls in hlg2genes.items():
        for genes in gene_ls:
            acc2gff_dict = acc2gff(gff, genes)
            comp_gff = []
            for acc, acc_gff in acc2gff_dict.items():
                comp_gff.extend(acc_gff)
            hlg2gffs[hlg].append(comp_gff)

    if gffs:
        for hlg, gffs in hlg2gffs.items():
            for i, gff in enumerate(gffs):
                if by_ome:
                    with open(f'{out_dir}{ome}/{hlg}.{i}.gff3', 'w') as out:
                        out.write(list2gff(gff))
                else:
                    with open(f'{out_dir}{hlg}/{ome}.{i}.gff3', 'w') as out:
                        out.write(list2gff(gff))
            
    if gbks:
        try:
            hlg2gbks = acc2gbk(ome, hlg2gffs, row)
        except KeyError:
            eprint(f'\tERROR: {ome} discrepancy between GFF and FAA', flush = True)
            hlg2gbks = {}
        for hlg, gbks in hlg2gbks.items():
            for i, gbk in enumerate(gbks):
                gbk_data = gbk[list(gbk.keys())[0]]
                if len(gbk) > 1:
                    eprint('\nERROR: multiple contigs for one cluster; huh?', flush = True)
                    sys.exit(9)
                if by_ome:
                    with open(f'{out_dir}{ome}/{hlg}.{i}.gbk', 'w') as out:
                        out.write(gbk_data)
                else:
                    with open(f'{out_dir}{hlg}/{ome}.{i}.gbk', 'w') as out:
                        out.write(gbk_data)
    
def compile_ome_faas(faa_path, ome, hlg2genes, out_dir, by_ome = False):
    faa = fa2dict(faa_path)
    for hlg, gene_ls in hlg2genes.items():
        for i, genes in enumerate(gene_ls):
            hlg_faa = {k: faa[k] for k in genes}
            if by_ome:
                with open(f'{out_dir}{ome}/{hlg}.{i}.faa', 'w') as out:
                    out.write(dict2fa(hlg_faa))
            else:
                with open(f'{out_dir}{hlg}/{ome}.{i}.faa', 'w') as out:
                    out.write(dict2fa(hlg_faa))


def main(db, clo_dir, out_dir, omes = [], hlgs = [], gcf_output = False, 
         make_gbk = False, make_gff = False, make_faa = False, cpus = 1):
    """Generate a dictionary of the output file formats"""

    db = db.set_index('ome')

    ome_dir = clo_dir + 'ome/'
    if gcf_output:
        prefix = 'gcf'
    else:
        prefix = 'hlg'
    if hlgs:
        print(f'\nParsing {prefix.upper()}s', flush = True)
        hlg_tsv = f'{clo_dir}{prefix}s.tsv.gz'
        hlg2omes, omes2hlg = parse_hlgtsv(hlg_tsv)
        with mp.Pool(processes = cpus) as pool:
            ome_res = pool.starmap(parse_ome, tqdm(((ome, f'{ome_dir}{ome}/{prefix}.tsv', hlgs) \
                                               for ome in omes2hlg), total = len(omes2hlg)))
        for hlg in hlgs:
            if not os.path.isdir(out_dir + str(hlg)):
                os.mkdir(out_dir + str(hlg))
    if omes:
        ome_set = set(omes)
        print(f'\nParsing ome {prefix.upper()}s', flush = True)
        ome_dirs = [x for x in os.listdir(ome_dir) \
                    if os.path.isdir(ome_dir + x) and x in db and x in ome_set]
        with mp.Pool(processes = cpus) as pool:
            ome_res = pool.starmap(parse_ome, tqdm(((ome, f'{ome_dir}{ome}/{prefix}.tsv') \
                                               for ome in ome_dirs), total = len(ome_dirs)))
        for ome in omes:
            if not os.path.isdir(out_dir + ome):
                os.mkdir(out_dir + ome)

    if make_gbk or make_gff:
        print(f'\nGenerating GFFs/GBKs', flush = True)
        if cpus > 1:
            with mp.Pool(processes = cpus) as pool:
                pool.starmap(compile_ome_gffs, 
                             tqdm(((db[ome], ome, hlg2genes, out_dir, bool(omes), make_gbk, make_gff) \
                              for ome, hlg2genes in ome_res if hlg2genes), total = len(ome_res)))
        else:
            for ome, hlg2genes in ome_res:
                if hlg2genes:
                    compile_ome_gffs(db[ome], ome, hlg2genes, out_dir, bool(omes), make_gbk, make_gff)
    if make_faa:
        print(f'\nGenerating protein fastas', flush = True)
        with mp.Pool(processes = cpus) as pool:
            pool.starmap(compile_ome_faas,
                         tqdm(((db[ome]['faa'], ome, hlg2genes, out_dir, bool(omes)) \
                              for ome, hlg2genes in ome_res), total = len(ome_res)))
        


def cli():
    parser = argparse.ArgumentParser(description = 'Parse CLOCI output and ' \
           + 'generate biofiles. "-" for stdin; -g specifies GCFs, all HLGs by default')

    in_opt = parser.add_argument_group('Input parameters')
    in_opt.add_argument('-c', '--cloci', help = 'CLOCI output dir', required = True)
    in_opt.add_argument('-g', '--gcf', action = 'store_true', help = 'Generate files for GCFs')
    in_opt.add_argument('-o', '--ome', help = 'Generate files for all ome(s) HLG/GCFs')
    in_opt.add_argument('--hlg', help = 'Generate files for specific HLG/GCF(s)')
    in_opt.add_argument('-d', '--mtdb', help = 'Reference MTDB; DEFAULT: primaryDB',
                        default = primaryDB())

    out_opt = parser.add_argument_group('Output parameters')
    out_opt.add_argument('--gbk', action = 'store_true', 
                         help = 'Generate cluster GenBank files')
    out_opt.add_argument('--gff', action = 'store_true',
                         help = 'Generate cluster GFF files')
    out_opt.add_argument('--faa', action = 'store_true',
                         help = 'Generate cluster amino acid fastas')
    out_opt.add_argument('--out', help = 'Output directory; DEFAULT: -c/biofiles/')

    run_opt = parser.add_argument_group('Runtime parameters')
    run_opt.add_argument('--cpu', type = int, default = 1)
    args = parser.parse_args()
    
    if len([x for x in [args.ome, args.hlg] if x]) > 1:
        eprint('\nERROR: -o and --hlg are incompatible', flush = True)
        sys.exit(4)

    if not args.gcf and not args.ome and not args.hlg:
        print('\nOutputting all HLGs', flush = True)
    elif args.gcf and not args.ome and not args.hlg:
        print('\nOutputting all GCFs', flush = True)


    run_dir = format_path(args.cloci)
    if not run_dir.endswith('/'):
        eprint('\nERROR: -c does not exist', flush = True)
        sys.exit(5)
    elif not args.gbk and not args.gff and not args.faa:
        eprint('\nERROR: --gbk/--gff/--faa required', flush = True)
        sys.exit(6)

    db = mtdb(args.mtdb)

    in_data = None
    for arg in [args.ome, args.hlg]:
        if arg:
            if '-' == arg:
                in_data = stdin2str()
            break

    out_dir = format_path(args.out)
    if not out_dir:
        out_dir = run_dir + 'biofiles/'
        if not os.path.isdir(out_dir):
            os.mkdir(out_dir)
    elif not out_dir.endswith('/'):
        os.mkdir(out_dir)
        out_dir += '/'

    omes, hlgs = [], []
    if args.ome:
        if in_data:
            omes = in_data.replace('"', '').replace("'", '').split().split(',')
        else:
            omes = [args.ome.replace('"', '').replace('"', '')]
        if args.gcf:
            print('\nOutputting GCFs for omes', flush = True)
        else:
            print('\nOutputting HLGs for omes', flush = True)
    elif args.hlg:
        if in_data:
            hlgs = [int(x) for x in in_data.replace('"', '').replace("'", '').split()]
        else:
            hlgs = [int(x) for x in args.hlg.replace('"', '').replace('"', '').split()]
        if args.gcf:
            print('\nOutputting specified GCFs', flush = True)
        else:
            print('\nOutputting specified HLGs', flush = True)

    main(db, run_dir, out_dir, omes, hlgs, args.gcf, 
         args.gbk, args.gff, args.faa, args.cpu)
    sys.exit(0)


if __name__ == '__main__':
    cli()



import os
import sys
import shutil
import random
import pickle
import multiprocessing as mp
from tqdm import tqdm
from cogent3 import PhyloNode
from itertools import combinations, chain
from collections import defaultdict, Counter
from mycotools.lib.biotools import gff2list
from mycotools.lib.kontools import eprint, collect_files
from cloci.lib import treecalcs
from cloci.lib.input_parsing import compileCDS

def hash_4_nulls_bysize(
    gff_path, gene2hg, size, plusminus, ome
    ):

    gff_list = gff2list(gff_path) # open here for decreased serialization
    cds_dict = compileCDS(gff_list, os.path.basename(gff_path).replace('.gff3',''))
  
    nulls = []
    for scaf, seqs in cds_dict.items():
        for i, seq in enumerate(seqs):
            if i < plusminus: # compile the locus
                locus = seqs[:i+plusminus+1]
            else:
                locus = seqs[i-plusminus:i+plusminus+1]
            loc_og = set([gene2hg[x] for x in locus if x in gene2hg]) # translate to OGs
            nulls.extend((
                [tuple(sorted(x)) for x in combinations(loc_og, size)]
                ))

    return ome, set(nulls)


def hash_4_nulls(
    gff_path, gene2hg, max_size, plusminus, ome = None
    ):

    gff_list = gff2list(gff_path) # open here for decreased serialization
    cds_dict = compileCDS(gff_list, os.path.basename(gff_path).replace('.gff3',''))
    sml_sizes = {i: [] for i in range(3, (plusminus * 2) + 2)} # for HGx sizes
    if max_size > (plusminus*2) + 1: # if biggest HGx is beyond the window size
        lrg_sizes = {i: [] for i in range((plusminus * 2) + 1, max_size + 1)}
        # create a separate dictionary to handle these
    else:
        lrg_sizes = {}

    for scaf, seqs in cds_dict.items():
        for i, seq in enumerate(seqs):
            if i < plusminus: # compile the locus
                locus = seqs[:i+plusminus+1]
            else:
                locus = seqs[i-plusminus:i+plusminus+1]
            loc_og = set([gene2hg[x] for x in locus if x in gene2hg]) # translate to OGs
            for size, size_list in sml_sizes.items():
                size_list.extend((
                    [tuple(sorted(x)) for x in combinations(loc_og, size)]
                    ))
        for size, size_list in lrg_sizes.items():
            size_list.extend([
                tuple(sorted(seqs[x:x+size])) for x in range(len(scaf)) \
                if len(seqs[x:x+size+1]) == size
                ])

    nulls = {
        **{k: set(v) for k,v in sml_sizes.items()},
        **{k: set(v) for k,v in lrg_sizes.items()}
        } # make them unique sets of combinations
    return ome, nulls


def form_cooccur_dict(cooccur_dict):

    # sort the values by the length of the combination
    cooccur_dict = {k: tuple(sorted(v)) for k, v in sorted(cooccur_dict.items(), key = lambda x: len(x[0]))}

    hgx2i, i2hgx = {}, {}
    for i, hgx in enumerate(list(cooccur_dict.keys())):
        i2hgx[i] = hgx
        hgx2i[hgx] = i

    return i2hgx, hgx2i, cooccur_dict


def gen_null_dict_omes(db, i2ome, combo_dict, sample = 10000, omes = []):
    """combo_dict = {ome: [(OG0...OGn)]}"""

    # make a list of observed HGx combinations 
    nulls = []
    for ome in omes:
        combo_dict[ome] = list(combo_dict[ome])
        nulls.extend(combo_dict[ome]) # extend all combos to the null
    
    # randomly sample while accounting for oversampled spp
    # NO DO NOT DO THIS. This approach was designed to account for sampling bias
    # in null distributions that have oversampled taxa, e.g. Saccharomyces, Cryptococcus
    # However, lineages that are not monophyletic in the microsynteny tree will have
    # a chance of small distances being enough to be declared unexpected because
    # the sampling bias wasn't considered.
    spp2omes, null_list = defaultdict(list), []
    [spp2omes[db[i2ome[x]]['taxonomy']['species']].append(x) for x in omes]
    spp = list(spp2omes.keys())
    sp_omes = list(chain(*list(spp2omes.values())))
    for i in range(sample):
#        rand_sp = random.choice(spp)
#        rand_ome = random.choice(spp2omes[rand_sp])
        rand_ome = random.choice(sp_omes)
        null_list.append(random.choice(combo_dict[rand_ome]))
        
    reps = {k: v for k, v in Counter(null_list).items() if v > 1}
    null_set = set(null_list)
                                            
    cooccur_dict = defaultdict(list)
    for ome, hgxs in combo_dict.items():
        intersect = list(set(hgxs).intersection(null_set))
        for hgx in intersect:
            cooccur_dict[hgx].append(ome)
    
    cooccur_dict = {k: tuple(v) for k, v in cooccur_dict.items() \
                    if len(v) > 1}
    # {(OG0,OGn): [omei0...]}
    reps = {k: v for k, v in reps.items() \
            if k in cooccur_dict}

    return cooccur_dict, reps


def gen_null_dict(combo_dict, sample = 10000):
    """combo_dict = {ome: [(OG0...OGn)]}"""
  
    nulls = []
    for combos in list(combo_dict.values()):
        nulls.extend(list(combos)) # extend all combos to the null
            
    if not isinstance(combo_dict[list(combo_dict.keys())[0]], set):
        combo_dict = {k: set(v) for k, v in combo_dict.items()}
        
            
    null_set_list = list(set(nulls)) # acquire unique combinations
    try: # sample
        null_list = random.sample(null_set_list, sample)
    except ValueError:
        print('\t\t\t\tWARNING: requested null sample greater than HGx size', flush = True)
        null_list = null_set_list
                                            
    cooccur_dict = {}                       
    for null in null_list: # for combo      
        cooccur_dict[null] = []             
        for ome, combos in combo_dict.items():
            if null in combos:
                cooccur_dict[null].append(ome)
    
    cooccur_dict = {x: cooccur_dict[x] for x in cooccur_dict}
    # {(OG0,OGn): [omei0...]}

    hgx2i, i2hgx, cooccur_dict = form_cooccur_dict(cooccur_dict)
    return hgx2i, i2hgx, cooccur_dict


def rank_partition(db, wrk_dir, rank, ome2i, partition_file):
    lineages = defaultdict(list)
    for ome, row in db.items():
        lineage = row['taxonomy'][rank]
        if lineage:
            if ome in ome2i:
                lineages[row['taxonomy'][rank]].append(ome)
    with open(partition_file, 'w') as out:
        for lineage, omes in lineages.items():
            out.write(f'#{lineage}\n{" ".join(omes)}\n')


def load_partitions(partition_file, ome2i):
    omes = set(ome2i.keys())
    with open(partition_file, 'r') as raw:
        partition_omes = [[ome2i[y] for y in x.rstrip().split() if y in omes] \
                          for x in raw \
                          if x.rstrip() and not x.startswith('#')]
    return partition_omes


def gen_nulls(db, i2ome, pairs, phylo, omes = [], samples = 10000, 
              uniq_sp = False, dist_func = treecalcs.calc_tmd,
              cpus = 1):
    """pairs = {ome: [(OG0,OG1)...]}"""

    null_dict, reps = gen_null_dict_omes(db, i2ome, pairs, samples, omes = omes)
    oldLen = len(null_dict)
    null_dict = {k: v for k, v in null_dict.items() if len(v) > 1}
    if not uniq_sp:
        results = treecalcs.calc_dists(phylo, null_dict, cpus = cpus,
                                       func = dist_func)
    else:
        results = treecalcs.calc_dists(phylo, null_dict, cpus = cpus, i2ome = i2ome,
                                       func = treecalcs.calc_tmd, uniq_sp = uniq_sp)
    omes2dist, pair_scores = {x[1]: x[0] for x in results}, []
    for k in null_dict:
        pair_scores.append(omes2dist[null_dict[k]])
    for k, iters in reps.items():
        for i in range(1, iters):
            pair_scores.append(omes2dist[null_dict[k]])
    for i in range(samples - len(pair_scores)):
        pair_scores.append(0)

    return omes2dist, sorted(pair_scores)


def gen_pair_nulls(db, phylo, ome2i, wrk_dir, nul_dir, seed_perc, ome2pairs,
                   i2ome, samples = 1000, partition_file = None, partition_rank = None,
                   dist_func = treecalcs.calc_tmd, uniq_sp = False, cpus = 1):
    if not os.path.isdir(nul_dir):
        os.mkdir(nul_dir)
    if partition_file or partition_rank:
        if partition_rank:
            partition_file = f'{wrk_dir}{partition_rank}_partitions.txt'
            if not os.path.isfile(partition_file):
                rank_partition(db, wrk_dir, partition_rank, ome2i, partition_file)
        print('\tPartitioning null distributions', flush = True)
        partition_omes = load_partitions(partition_file, ome2i)
        ome2partition = {}
        for k, omes in enumerate(partition_omes):
            for ome in omes:
                ome2partition[ome] = k
        missing_omes = set(ome2i.values()).difference(set(ome2partition.keys()))
        ome2partition = {**ome2partition, **{k: None for k in missing_omes}}
    else: # spoof it
        partition_omes = [ome2i.values()]
        ome2partition = {i: 0 for i in ome2i.values()}

    if os.path.isfile(wrk_dir + 'omes2dist.pickle'):
        with open(wrk_dir + 'omes2dist.pickle', 'rb') as raw:
            omes2dist = pickle.load(raw)
    else:
        omes2dist = {}

    if not os.path.isfile(f'{wrk_dir}hgx2loc.pickle'):
        # create null distribution for orthogroup pairs
        final_partition = len(partition_omes) - 1 
        pair_nulls, omes2dist = [], {}
        if not os.path.isfile(f'{nul_dir}2.null.{final_partition}.txt'):
            print('\tGenerating null distributions', flush = True)
            for i, omes in enumerate(partition_omes):
                omes2dist, pair_null = gen_nulls(
                    db, i2ome, ome2pairs, phylo,
                    omes, samples = samples, 
                    dist_func = dist_func, uniq_sp = uniq_sp,
                    cpus = cpus
                    )
                n_f = f'{nul_dir}2.null.{i}.txt'
                with open(f'{n_f}.tmp', 'w') as out:
                    out.write('\n'.join([str(i0) for i0 in pair_null]))
                shutil.move(f'{n_f}.tmp', n_f)
                pair_nulls.append(pair_null)
            with open(wrk_dir + 'omes2dist.pickle', 'wb') as out:
                pickle.dump(omes2dist, out)
        else: # or just load what is available
            if os.path.isfile(wrk_dir + 'omes2dist.pickle'):
                with open(wrk_dir + 'omes2dist.pickle', 'rb') as raw:
                    omes2dist = pickle.load(raw)
            for i, omes in enumerate(partition_omes):
                with open(f'{nul_dir}2.null.{i}.txt', 'r') as raw:
                   pair_nulls.append([float(x.rstrip()) for x in raw])
        min_pair_scores = []
        for i, pair_null in enumerate(pair_nulls):
            scores_i = round(seed_perc * len(pair_null) + .5)
            min_pair_scores.append(pair_null[scores_i])
    else:
        min_pair_scores = None

    return partition_omes, ome2partition, \
           omes2dist, min_pair_scores


def load_null(nul_f):
    with open(nul_f, 'r') as raw:
        nullSizes = [float(x.rstrip()) for x in raw]
    nullSizes.sort()
    return nullSizes


def load_hgx_nulls(max_clus_size, nul_dir, hgx_perc, clus_perc, partition_num):

    hgxBordPercs, hgxClusPercs = {}, {}
    for size in range(3, max_clus_size + 1):
        nul_f = f'{nul_dir}{size}.null.{partition_num}.txt'
        if os.path.isfile(nul_f):
            with open(nul_f, 'r') as raw:
                nullSizes = [float(x.rstrip()) for x in raw]
            nullSizes.sort()
            hgxBordPercs[size] = nullSizes[round(hgx_perc * len(nullSizes) + .5)]
            hgxClusPercs[size] = nullSizes[round(clus_perc * len(nullSizes) + .5)]
        elif os.path.isfile(f'{nul_dir}skip.{size}-{partition_num}.txt'):
            hgxBordPercs[size] = 0.0
            hgxClusPercs[size] = 0.0
        else:
            raise FileNotFoundError(f'missing null/skip file in {nul_dir}')
    
    return hgxBordPercs, hgxClusPercs


def parse_done_file(done_file):
    # NEED some accounting of changing bord percentile
    skip_i = {}
    done_data = done_file.split('.')
    if done_data:
        for d in done_data[1:]:
            k,v = d.split('-')
            skip_i[int(k)] = int(v)
    return skip_i


def parse_skip_files(skip_files):
    # NEED some accounting of changing bord percentile
    skip_i = defaultdict(int)
    for skip_f in skip_files:
        s, i = [int(x) for x in skip_f.replace('skip.','').replace('.txt','').split('-')]
        if s > skip_i[i]:
            skip_i[i] = s
    return {k: v for k,v in skip_i.items()}


def gen_hgx_nulls(
    db, i2ome, gffs, gene2hg, max_clus_size, plusminus, phylo,
    hgx_perc, clus_perc, nul_dir, partition_omes,
    omes2dist = {}, skip_i = {}, samples = 10000, 
    dist_func = treecalcs.calc_tmd, uniq_sp = False, cpus = 1
    ):
    """Generates a null distribution of randomly sampled HGxs for each 
    # of homogroups observed in HGxs. Applies the percentile for each
    size as the minimum value for significance.
    Outputs a dictionary of the minimum values for each size HGx
    based on the inputted percentiles. {# of OGs: minimum value}"""

    # hashRes = [({size: [HGx]})...] by ome
    print('\tCalculating random sample TMD', flush = True)
    hgxBordPercs, hgxClusPercs = defaultdict(dict), defaultdict(dict)
    for size in range(3, max_clus_size + 1): # for all observed sizes
        if all(i in skip_i for i, v in enumerate(partition_omes)):
            if all(skip_i[i] <= size for i, v in enumerate(partition_omes)):
                for i, v in enumerate(partition_omes):
                    hgxBordPercs[i][size] = 0.0
                    hgxClusPercs[i][size] = 0.0
                continue
        print(f'\t\tHGx size {size}', flush = True)
        if not all(os.path.isfile(f'{nul_dir}{size}.null.{i}.txt') \
               or os.path.isfile(f'{nul_dir}skip.{size}-{i}') \
               for i in range(len(partition_omes))):
            print('\t\t\tRandomly sampling HGxs', flush = True)
            hash_null_cmds = [
                (x, gene2hg, size, plusminus, i) \
                for x, i in gffs.items()
                ]
            with mp.get_context('fork').Pool(processes = cpus) as pool:
                hashRes = pool.starmap(hash_4_nulls_bysize, 
                                       tqdm(hash_null_cmds, total = len(hash_null_cmds)))
                pool.close()
                pool.join() 
            size_dict = {ome: nulls for ome, nulls in hashRes}
            size_dict = {k: size_dict[k] for k in sorted(size_dict.keys())}
        for i0, omes in enumerate(partition_omes):
            if i0 in skip_i:
                if skip_i[i0] <= size:
                    hgxBordPercs[i0][size] = 0.0
                    hgxClusPercs[i0][size] = 0.0
                    continue
            # multiprocessing
            if os.path.isfile(f'{nul_dir}{size}.null.{i0}.txt'):
                nullSizes = load_null(f'{nul_dir}{size}.null.{i0}.txt')
            else:
                null_dict, reps = gen_null_dict_omes(db, i2ome, size_dict, samples, omes)
                omes2dist = treecalcs.update_dists(phylo, null_dict, func = dist_func,
                                                   uniq_sp = uniq_sp, i2ome = i2ome,
                                                   omes2dist = omes2dist, cpus = cpus)
                nullSizes = []
                with open(f'{nul_dir}{size}.null.{i0}.txt.tmp', 'w') as out:
                    for omes in list(null_dict.values()):
                        nullSizes.append(omes2dist[omes])
                        out.write(str(omes2dist[omes]) + '\n')
                    count = 0
                    for k, v in reps.items():
                        for i1 in range(v - 1):
                            t_dist = omes2dist[null_dict[k]]
                            out.write(str(t_dist) + '\n')
                            count += 1
                            nullSizes.append(t_dist)
                    z_vals = range(samples - len(null_dict) - count)
                    out.write('\n'.join([str(0) for x in z_vals]))
                    nullSizes.extend([0 for x in z_vals])
                    # account for missing values
                shutil.move(f'{nul_dir}{size}.null.{i0}.txt.tmp',
                            f'{nul_dir}{size}.null.{i0}.txt')
        
            # sort to apply percentile threshold for each size HGx and for both the
            # border percentile and cluster percentile
                nullSizes.sort()
            hgxBordPercs[i0][size] = nullSizes[round(hgx_perc * len(nullSizes) + .5)]
            hgxClusPercs[i0][size] = nullSizes[round(clus_perc * len(nullSizes) + .5)]
            if not hgxBordPercs[i0][size]:
                skip_i[i0] = size
                for s in range(size, max_clus_size + 1):
                    with open(f'{nul_dir}skip.{s}-{i0}.txt', 'w') as out:
                        pass
                print(f'\t\t\tWARNING: lineage {i0}: all HGx >= {size} HGs pass')

    #skips = [f'{k}-{v}' for k,v in skip_i.items()]
#    done_file = f'{nul_dir}done.{".".join(skips)}'
    done_file = f'{nul_dir}done.txt'
    with open(done_file, 'w') as out:
        pass

    return hgxBordPercs, hgxClusPercs


def partitions2hgx_nulls(db, partition_omes, ome2i, i2ome, gene2hg, max_hgx_size,
                         plusminus, hgx_perc, clus_perc, nul_dir, omes2dist,
                         phylo, samples = 1000, dist_func = treecalcs.calc_tmd,
                         uniq_sp = False, cpus = 1):
    final_partition = len(partition_omes) - 1
    print('\tPreparing HGx nulls', flush = True)
    files = [os.path.basename(x) for x in collect_files(nul_dir, 'txt')]
    done_file = [x for x in files if x.startswith('done.')]
    skip_files = [x for x in files if x.startswith('skip.')]
    skip_i = parse_skip_files(skip_files)
    if done_file:
        bordScores_list, clusScores_list = [], []
#        skip_i = parse_done_file(done_file)
        for i, omes in enumerate(partition_omes):
            bordScores, clusScores = load_hgx_nulls(max_hgx_size, nul_dir, hgx_perc,
                                                    clus_perc, i)
            bordScores_list.append(bordScores)
            clusScores_list.append(clusScores)
    else:
        bordScores_list, clusScores_list = gen_hgx_nulls(
            db, i2ome, {db[k]['gff3']: ome2i[k] for k in list(db.keys()) \
             if k in ome2i},
            gene2hg, max_hgx_size, plusminus, phylo,
            hgx_perc, clus_perc, nul_dir, partition_omes,
            omes2dist, skip_i, samples = samples, cpus = cpus,
            dist_func = dist_func, uniq_sp = uniq_sp
            )

    return bordScores_list, clusScores_list



import os
import re
import random
import pickle
import multiprocessing as mp
from tqdm import tqdm
from cogent3 import PhyloNode
from itertools import chain
from collections import Counter, defaultdict
from mycotools.lib.kontools import eprint

def addPatch(phylo, omes_set):
    """recursively add branches with the trait to the presence total"""
    total = 0
    for sphylo in phylo:
        subOmes = [str(x)[:str(x).find(':')] for x in sphylo.tips(True)]
        if any(
            x in omes_set \
            for x in subOmes
            ): # if some descendents aren't in the set, search this branch
            if all(x in omes_set for x in subOmes):
                if len(subOmes) > 1:
                    total += sphylo.total_descending_branch_length()
                else:
                    total += sphylo.length
            else:
                total += addPatch(sphylo, omes_set)
    return total

def calc_pds(phylo, omes):
    """calculate the percent branch length that a trait is missing over the
    MRCA of where the trait is present"""
    omes_set = set(omes)
    try:
        mrca = phylo.lowest_common_ancestor(omes)
 #   except ValueError:
#        print(phylo, omes)
    except AttributeError: # 1 ome
        eprint('\t\t' + ','.join([str(x) for x in omes]) \
             + ' missing tip(s)', flush = True)
        print(phylo)

    mrca_omes = set(x.name for x in mrca.iter_tips())
    if mrca_omes == omes_set:
        return tuple([int(x) for x in omes]), 0
    else:
        totalDist = mrca.total_descending_branch_length()
        subDist = addPatch(mrca, omes_set)
        return tuple([int(x) for x in omes]), 1 - subDist/totalDist



def patch_main(
    phylo, omes, wrk_dir,
    old_path = 'pds.pickle', cpus = 1
    ):

    if os.path.isfile(wrk_dir + old_path):
        print('\tLoading previous PDS results', flush = True)
        with open(wrk_dir + old_path, 'rb') as in_pick:
            omes2patch = pickle.load(in_pick)
    else:
        omes2patch = {}

    clusOmes = set([
        tuple([str(x) for x in y]) for y in omes \
               if y not in omes2patch
        ])
    if clusOmes:
        with mp.get_context('fork').Pool(processes = cpus) as pool:
            patch_res = pool.starmap(
                calc_pds, tqdm([(phylo, x) for x in clusOmes],
                                                 total = len(clusOmes))
                )
            pool.close()
            pool.join()
        omes2patch = {**omes2patch,
                      **{ome_tup: pds for ome_tup, pds in patch_res}}
        with open(wrk_dir + old_path, 'wb') as out:
            pickle.dump(omes2patch, out)
    
    return omes2patch

def calc_mmd(phylo, omes):
    mrca = phylo.lowest_common_ancestor([str(x) for x in omes])
    mmd = mrca.get_max_tip_tip_distance()[0]
    return mmd, tuple([int(x) for x in omes])

def calc_tmd(phylo, omes, error = False):
    """calculate descending branch length from cogent3 tree"""
    # need to verify wtf descending branch length v total of supplied nodes is
    omes = [str(x) for x in omes]
    omes_set = set(omes)

    try: 
        mrca = phylo.lowest_common_ancestor(omes) # this is the failing step
        mrca_omes = set(x.name for x in mrca.iter_tips())
        tmd = addPatch(phylo.lowest_common_ancestor(omes),
                       omes_set)
        return tmd, tuple([int(i) for i in omes])
    except ValueError:
        # get the missing tips
        missing_tips = set(omes).difference(set(phylo.get_tip_names()))
        # get the extraneous tips if no missing
        if not missing_tips:
            extraneous_tips = set(phylo.get_tip_names()).difference(set(omes))
            if error:
                eprint(f'\t\tERROR: {",".join(sorted(extraneous_tips))}' \
                     + ' extraneous tip(s)', flush = True)
                sys.exit(13)
            else:
                eprint(f'\t\tWARNING: {",".join(sorted(extraneous_tips))}' \
                     + ' extraneous tip(s)', flush = True)
        else:
            if error:
                eprint(f'\t\tERROR: {",".join(sorted(missing_tips))}' \
                     + ' missing tip(s)', flush = True)
                sys.exit(13)
            else:
                eprint(f'\t\tWARNING: {",".join(sorted(missing_tips))}' \
                     + ' missing tip(s)', flush = True)

        return 0, tuple([int(i) for i in omes])
    except AttributeError:
        print(omes, '\n', phylo)

def calc_tmd_uniq_omes(phylo, omes, o_omes):
    """calculate descending branch length from cogent3 tree"""
    # need to verify wtf descending branch length v total of supplied nodes is
    omes = [str(x) for x in omes]
    omes_set = set(omes)
    try:
        mrca = phylo.lowest_common_ancestor(omes)
        mrca_omes = set(x.name for x in mrca.iter_tips())
        tmd = addPatch(phylo.lowest_common_ancestor(omes),
                       omes_set)
        return tmd, tuple([int(i) for i in o_omes])
    except ValueError:
        eprint('\t\t' + ','.join([str(x) for x in omes]) \
             + ' missing/extraneous tip(s)', flush = True)
        return 0, tuple([int(i) for i in o_omes])
    except AttributeError:
        print(omes, '\n', phylo)


def calc_branch_sim(phylo, omes0, omes1):
    tmd_union = calc_tmd(phylo, list(set(omes0).union(set(omes1))))[0]
    tmd_inter = calc_tmd(phylo, list(set(omes0).intersection(set(omes1))))[0]
    return tmd_inter/tmd_union


def get_uniq_spp(db, iomes, i2ome):
    omes = {i2ome[x]: x for x in iomes}
    ab_omes = [re.search(r'([^\d+])\d', x)[1] for x in omes]
    reps_set = set(k for k, v in Counter(ab_omes).items() if v > 1)
    # if there are duplicate omes, then check if the species is duplicate
    if reps_set:
        rep_spp = defaultdict(list)
        [rep_spp[db[x]['taxonomy']['species']].append(x) for x in omes \
         if re.search(r'([^\d+])\d', x)[1] in reps_set]
        # need the omes that weren't duplicated, the ones that were but aren't the 
        # same species, and a random choice of the duplicated ones
        # species and their omes that were duplicated
        dup_spp = {k: v for k, v in rep_spp.items() if len(v) > 1}
        dup_omes = chain(*list(dup_spp.values()))
        uniq_omes = list(set(omes.keys()).difference(set(dup_omes)))
        for sp, rep_omes in dup_spp.items():
            r_ome = rep_omes[random.randint(0, len(rep_omes) - 1)]
            uniq_omes.append(r_ome)
        return tuple(sorted([omes[i] for i in uniq_omes]))
    else:
        return iomes
        

def calc_dists(phylo, cooccur_dict, cpus = 1, omes2dist = {}, func = calc_tmd,
               uniq_sp = False, i2ome = None):
    # multiprocessing calculating only new distances for omes2dist
    if uniq_sp:
        with mp.get_context('forkserver').Pool(processes = cpus) as pool:
            results = pool.starmap(
                calc_tmd_uniq_omes,
                [(phylo, get_uniq_spp(uniq_sp, x, i2ome), x) \
                  for x in list(set(cooccur_dict.values())) \
                  if x not in omes2dist]
                )
            pool.close()
            pool.join()
    else:
        with mp.get_context('forkserver').Pool(processes = cpus) as pool:
            results = pool.starmap(
                func,
                [(phylo, x,) for x in list(set(cooccur_dict.values())) \
                if x not in omes2dist]
                )
            pool.close()
            pool.join()
        
    return results 


def update_dists(phylo, cooccur_dict, cpus = 1, omes2dist = {}, func = calc_tmd,
                 uniq_sp = [], i2ome = None):
    """update the omes2dist with a new set of distance calulations"""
    results = calc_dists(phylo, cooccur_dict, cpus, omes2dist = omes2dist, func = func,
                         uniq_sp = uniq_sp, i2ome = i2ome)
    omes2dist = {**omes2dist, **{x[1]: x[0] for x in results}}
    return omes2dist
